import numpy as np

from tqdm import tqdm
import scipy.io

from dipy.io.stateful_tractogram import StatefulTractogram, Space
import nibabel as nib

from AFQ.data.utils import BUNDLE_MAT_2_PYTHON


# This dictionary is used to convert the names of the bundles
# from old pyAFQ bundle names to the new pyAFQ bundle names.
old_acronyms_to_formal = {
    "ATR_L": "Left Anterior Thalamic",
    "ATR_R": "Right Anterior Thalamic",
    "CST_L": "Left Corticospinal",
    "CST_R": "Right Corticospinal",
    "CGC_L": "Left Cingulum Cingulate",
    "CGC_R": "Right Cingulum Cingulate",
    "IFO_L": "Left Inferior Fronto-Occipital",
    "IFO_R": "Right Inferior Fronto-Occipital",
    "ILF_L": "Left Inferior Longitudinal",
    "ILF_R": "Right Inferior Longitudinal",
    "SLF_L": "Left Superior Longitudinal",
    "SLF_R": "Right Superior Longitudinal",
    "UNC_L": "Left Uncinate",
    "UNC_R": "Right Uncinate",
    "ARC_L": "Left Arcuate",
    "ARC_R": "Right Arcuate",
    "VOF_L": "Left Vertical Occipital",
    "VOF_R": "Right Vertical Occipital",
    "pARC_L": "Left Posterior Arcuate",
    "pARC_R": "Right Posterior Arcuate",
    "Orbital": "Callosum Orbital",
    "AntFrontal": "Callosum Anterior Frontal",
    "SupFrontal": "Callosum Superior Frontal",
    "Motor": "Callosum Motor",
    "SupParietal": "Callosum Superior Parietal",
    "PostParietal": "Callosum Posterior Parietal",
    "Occipital": "Callosum Occipital",
    "Temporal": "Callosum Temporal"
}


class MatlabFileTracking():
    """
    Helper class.
    Acts the same as a tracking class from DIPY,
    in that it yields a streamline for each call to __iter__.
    Initialized with an opened h5py matlab file
    and the location of the streamlines in that h5py file.
    """

    def __init__(self, fg_ref):
        self.fg_ref = fg_ref

    def __iter__(self):
        for i in tqdm(range(self.fg_ref.shape[0])):
            yield np.transpose(self.fg_ref[i, 0])


def matlab_tractography(mat_file, img):
    """
    Converts a matlab tractography file to a stateful tractogram.

    Parameters
    ----------
    mat_file : str
        Path to a matlab tractography file.
    img : Nifti1Image or str
        Path to an img file to be loaded with nibabel or an img
        to serve as the reference for the stateful tractogram.

    Returns
    -------
    DIPY :class:`StatefulTractogram` in RASMM space.
    """
    mat_file = scipy.io.loadmat(mat_file)
    if isinstance(img, str):
        img = nib.load(img)

    tracker = MatlabFileTracking(mat_file['fg']['fibers'][0][0])
    return StatefulTractogram(tracker, img, Space.RASMM)


def matlab_mori_groups(mat_file, img):
    """
    Converts a matlab Mori groups file to a dictionary of fiber groups.
    This dictionary is structured the same way as the results of pyAFQ
    segmentation. The keys are bundle names and the values are 
    :class:`StatefulTractogram` instances.
    If you want to merge this dictionary into one :class:`StatefulTractogram`,
    use :class:`SegmentedSFT`.

    Parameters
    ----------
    mat_file : str
        Path to a matlab Mori groups file.
    img : Nifti1Image or str
        Path to an img file to be loaded with nibabel or an img
        to serve as the reference for the stateful tractogram.

    Returns
    -------
    Dictionary where keys are the pyAFQ bundle names and values are
    DIPY :class:`StatefulTractogram` instances in RASMM space.
    """
    mat_file = scipy.io.loadmat(mat_file)
    if isinstance(img, str):
        img = nib.load(img)

    fiber_groups = {}
    for i in range(mat_file["fg"]["name"].shape[1]):
        name = mat_file["fg"]["name"][0][i][0]
        py_name = None
        if name in BUNDLE_MAT_2_PYTHON.keys():
            py_name = BUNDLE_MAT_2_PYTHON[name]
        elif name in BUNDLE_MAT_2_PYTHON.values():
            py_name = name
        if py_name is not None:
            bundle_ref = mat_file["fg"]["fibers"][0][i]
            tracker = MatlabFileTracking(bundle_ref)
            fiber_groups[py_name] =\
                StatefulTractogram(tracker, img, Space.RASMM)

    return fiber_groups



import nibabel as nib
import os
import os.path as op
from time import time
import numpy as np
import pandas as pd
import logging

import pimms

from AFQ.tasks.decorators import as_file
from AFQ.tasks.utils import get_fname, with_name, str_to_desc
from AFQ.recognition.recognize import recognize
from AFQ.utils.path import drop_extension, write_json
import AFQ.utils.streamlines as aus
from AFQ.tasks.utils import get_default_args
import AFQ.utils.volume as auv
from AFQ._fixes import gaussian_weights

try:
    from trx.io import load as load_trx
    from trx.io import save as save_trx
    from trx.trx_file_memmap import TrxFile
    has_trx = True
except ModuleNotFoundError:
    has_trx = False

from dipy.io.streamline import load_tractogram, save_tractogram
from dipy.io.stateful_tractogram import Space
from dipy.stats.analysis import afq_profile
from dipy.tracking.streamline import set_number_of_points, values_from_volume


logger = logging.getLogger('AFQ')


@pimms.calc("bundles")
@as_file('_tractography', include_track=True, include_seg=True)
def segment(data_imap, mapping_imap,
            tractography_imap, segmentation_params):
    """
    full path to a trk/trx file containing containing
    segmented streamlines, labeled by bundle

    Parameters
    ----------
    segmentation_params : dict, optional
        The parameters for segmentation.
        Default: use the default behavior of the seg.Segmentation object.
    """
    bundle_dict = data_imap["bundle_dict"]
    reg_template = data_imap["reg_template"]
    streamlines = tractography_imap["streamlines"]
    if streamlines.endswith(".trk") or streamlines.endswith(".tck"):
        tg = load_tractogram(
            streamlines, data_imap["dwi"], Space.VOX,
            bbox_valid_check=False)
        is_trx = False
    elif streamlines.endswith(".trx"):
        is_trx = True
        trx = load_trx(streamlines, data_imap["dwi"])
        trx.streamlines._data = trx.streamlines._data.astype(np.float32)
        tg = trx.to_sft()

    indices_to_remove, _ = tg.remove_invalid_streamlines()
    if len(indices_to_remove) > 0:
        logger.warning(f"{len(indices_to_remove)} invalid streamlines removed")

    start_time = time()
    bundles, bundle_meta = recognize(
        tg,
        data_imap["dwi"],
        mapping_imap["mapping"],
        bundle_dict,
        reg_template,
        **segmentation_params)

    seg_sft = aus.SegmentedSFT(bundles, Space.VOX)

    if len(seg_sft.sft) < 1:
        raise ValueError("Fatal: No bundles recognized.")

    if is_trx:
        seg_sft.sft.dtype_dict = {'positions': np.float16,
                                  'offsets': np.uint32}
        tgram = TrxFile.from_sft(seg_sft.sft)
        tgram.groups = seg_sft.bundle_idxs
        meta = {}

    else:
        tgram, meta = seg_sft.get_sft_and_sidecar()

    seg_params_out = {}
    for arg_name, value in segmentation_params.items():
        if isinstance(value, (int, float, bool, str)):
            seg_params_out[arg_name] = value
        elif isinstance(value, (list, tuple)):
            seg_params_out[arg_name] = [str(v) for v in value]
        elif isinstance(value, dict):
            for k, v in value.items():
                seg_params_out[k] = str(v)
        else:
            seg_params_out[arg_name] = str(value)

    meta["source"] = streamlines
    meta["Recognition Parameters"] = seg_params_out
    meta["Bundle Parameters"] = bundle_meta
    meta["Timing"] = time() - start_time
    return tgram, meta


@pimms.calc("indiv_bundles")
def export_bundles(base_fname, output_dir,
                   bundles,
                   tracking_params,
                   segmentation_params):
    """
    dictionary of paths, where each path is
    a full path to a trk file containing the streamlines of a given bundle.
    """
    is_trx = tracking_params.get("trx", False)
    if is_trx:
        extension = ".trx"
    else:
        extension = ".trk"

    bundles_dir = op.join(output_dir, "bundles")
    os.makedirs(bundles_dir, exist_ok=True)
    seg_sft = aus.SegmentedSFT.fromfile(bundles)
    for bundle in seg_sft.bundle_names:
        fname = op.split(
            get_fname(
                base_fname,
                f'_desc-{str_to_desc(bundle)}'
                f'_tractography{extension}',
                tracking_params=tracking_params,
                segmentation_params=segmentation_params))
        fname = op.join(bundles_dir, fname[1])
        bundle_sft = seg_sft.get_bundle(bundle)
        if len(bundle_sft) > 0:
            logger.info(f"Saving {fname}")
            if is_trx:
                seg_sft.sft.dtype_dict = {
                    'positions': np.float16,
                    'offsets': np.uint32}
                trxfile = TrxFile.from_sft(bundle_sft)
                save_trx(trxfile, fname)
            else:
                save_tractogram(
                    bundle_sft, fname,
                    bbox_valid_check=False)
        else:
            logger.info(f"No bundle to save for {bundle}")
        meta = dict(
            source=bundles,
            params=seg_sft.get_bundle_param_info(bundle))
        meta_fname = drop_extension(fname) + '.json'
        write_json(meta_fname, meta)
    return bundles_dir


@pimms.calc("sl_counts")
@as_file('_desc-slCount_dwi.csv', include_track=True, include_seg=True)
def export_sl_counts(bundles):
    """
    full path to a JSON file containing streamline counts
    """
    sl_counts = []
    seg_sft = aus.SegmentedSFT.fromfile(bundles)

    for bundle in seg_sft.bundle_names:
        sl_counts.append(len(
            seg_sft.get_bundle(bundle).streamlines))
    sl_counts.append(len(seg_sft.sft.streamlines))

    counts_df = pd.DataFrame(
        data=dict(
            n_streamlines=sl_counts),
        index=seg_sft.bundle_names + ["Total Recognized"])
    return counts_df, dict(source=bundles)


@pimms.calc("median_bundle_lengths")
@as_file(
    '_desc-medianBundleLengths_dwi.csv',
    include_track=True, include_seg=True)
def export_bundle_lengths(bundles):
    """
    full path to a JSON file containing median bundle lengths
    """
    med_len_counts = []
    seg_sft = aus.SegmentedSFT.fromfile(bundles)

    for bundle in seg_sft.bundle_names:
        these_lengths = seg_sft.get_bundle(
            bundle)._tractogram._streamlines._lengths
        if len(these_lengths) > 0:
            med_len_counts.append(np.median(
                these_lengths))
        else:
            med_len_counts.append(0)
    med_len_counts.append(np.median(
        seg_sft.sft._tractogram._streamlines._lengths))

    counts_df = pd.DataFrame(
        data=dict(
            median_len=med_len_counts),
        index=seg_sft.bundle_names + ["Total Recognized"])
    return counts_df, dict(source=bundles)


@pimms.calc("density_maps")
@as_file('_desc-density_dwi.nii.gz', include_track=True, include_seg=True)
def export_density_maps(bundles, data_imap):
    """
    full path to 4d nifti file containing streamline counts per voxel
    per bundle, where the 4th dimension encodes the bundle
    """
    seg_sft = aus.SegmentedSFT.fromfile(
        bundles)
    entire_density_map = np.zeros((
        *data_imap["data"].shape[:3],
        len(seg_sft.bundle_names)))
    for ii, bundle_name in enumerate(seg_sft.bundle_names):
        bundle_sl = seg_sft.get_bundle(bundle_name)
        bundle_density = auv.density_map(bundle_sl).get_fdata()
        entire_density_map[..., ii] = bundle_density

    return nib.Nifti1Image(
        entire_density_map, data_imap["dwi_affine"]), dict(
            source=bundles, bundles=list(seg_sft.bundle_names))


@pimms.calc("profiles")
@as_file('_desc-profiles_dwi.csv', include_track=True, include_seg=True)
def tract_profiles(bundles,
                   scalar_dict, data_imap,
                   profile_weights="gauss",
                   n_points_profile=100):
    """
    full path to a CSV file containing tract profiles

    Parameters
    ----------
    profile_weights : str, 1D array, 2D array callable, optional
        How to weight each streamline (1D) or each node (2D)
        when calculating the tract-profiles. If callable, this is a
        function that calculates weights. If None, no weighting will
        be applied. If "gauss", gaussian weights will be used.
        If "median", the median of values at each node will be used
        instead of a mean or weighted mean.
        Default: "gauss"
    n_points_profile : int, optional
        Number of points to resample each streamline to before
        calculating the tract-profiles.
        Default: 100
    """
    if not (profile_weights is None
            or isinstance(profile_weights, str)
            or callable(profile_weights)
            or hasattr(profile_weights, "__len__")):
        raise TypeError(
            "profile_weights must be string, None, callable, or"
            + "a 1D or 2D array")
    if isinstance(profile_weights, str):
        profile_weights = profile_weights.lower()
    if isinstance(profile_weights, str) and\
            profile_weights != "gauss" and profile_weights != "median":
        raise TypeError(
            "if profile_weights is a string,"
            + " it must be 'gauss' or 'median'")

    bundle_names = []
    node_numbers = []
    profiles = np.empty((len(scalar_dict), 0)).tolist()
    this_profile = np.zeros((len(scalar_dict), n_points_profile))
    reference = nib.load(scalar_dict[list(scalar_dict.keys())[0]])
    seg_sft = aus.SegmentedSFT.fromfile(
        bundles,
        reference=reference)

    seg_sft.sft.to_rasmm()
    for bundle_name in seg_sft.bundle_names:
        this_sl = seg_sft.get_bundle(bundle_name).streamlines
        if len(this_sl) == 0:
            continue
        if profile_weights == "gauss":
            # calculate only once per bundle
            bundle_profile_weights = gaussian_weights(
                this_sl,
                n_points=n_points_profile)
        for ii, (scalar, scalar_file) in enumerate(scalar_dict.items()):
            if isinstance(scalar_file, str):
                scalar_file = nib.load(scalar_file)
            scalar_data = scalar_file.get_fdata()
            if isinstance(profile_weights, str):
                if profile_weights == "gauss":
                    this_prof_weights = bundle_profile_weights
                elif profile_weights == "median":
                    # weights bundle to only return the mean
                    def _median_weight(bundle):
                        fgarray = set_number_of_points(
                            bundle, n_points_profile)
                        values = np.array(
                            values_from_volume(
                                scalar_data,
                                fgarray,
                                data_imap["dwi_affine"]))
                        weights = np.zeros(values.shape)
                        for ii, jj in enumerate(
                            np.argsort(values, axis=0)[
                                len(values) // 2, :]):
                            weights[jj, ii] = 1
                        return weights
                    this_prof_weights = _median_weight
            else:
                this_prof_weights = profile_weights
            this_profile[ii] = afq_profile(
                scalar_data,
                this_sl,
                data_imap["dwi_affine"],
                weights=this_prof_weights,
                n_points=n_points_profile)
            profiles[ii].extend(list(this_profile[ii]))
        nodes = list(np.arange(this_profile[0].shape[0]))
        bundle_names.extend([bundle_name] * len(nodes))
        node_numbers.extend(nodes)

    profile_dict = dict()
    profile_dict["tractID"] = bundle_names
    profile_dict["nodeID"] = node_numbers
    for ii, scalar in enumerate(scalar_dict.keys()):
        profile_dict[scalar] = profiles[ii]

    profile_dframe = pd.DataFrame(profile_dict)
    meta = dict(source=bundles,
                parameters=get_default_args(afq_profile),
                scalars=list(scalar_dict.keys()),
                bundles=list(seg_sft.bundle_names))

    return profile_dframe, meta


@pimms.calc("scalar_dict")
def get_scalar_dict(data_imap, mapping_imap, scalars=["dti_fa", "dti_md"]):
    """
    dicionary mapping scalar names
    to their respective file paths

    Parameters
    ----------
    scalars : list of strings and/or scalar definitions, optional
        List of scalars to use.
        Can be any of: "dti_fa", "dti_md", "dki_fa", "dki_md", "dki_awf",
        "dki_mk". Can also be a scalar from AFQ.definitions.image.
        Default: ["dti_fa", "dti_md"]
    """
    # Note: some scalars preprocessing done in plans, before this step
    scalar_dict = {}
    for scalar in scalars:
        if isinstance(scalar, str):
            sc = scalar.lower()
            scalar_dict[sc] = data_imap[f"{sc}"]
        elif f"{scalar.get_name()}" in mapping_imap:
            scalar_dict[scalar.get_name()] = mapping_imap[
                f"{scalar.get_name()}"]
    return {"scalar_dict": scalar_dict}


def get_segmentation_plan(kwargs):
    if "segmentation_params" in kwargs\
            and not isinstance(kwargs["segmentation_params"], dict):
        raise TypeError(
            "segmentation_params a dict")
    segmentation_tasks = with_name([
        get_scalar_dict,
        export_sl_counts,
        export_bundle_lengths,
        export_bundles,
        export_density_maps,
        segment,
        tract_profiles])

    default_seg_params = get_default_args(recognize)
    if "segmentation_params" in kwargs:
        for k in kwargs["segmentation_params"]:
            default_seg_params[k] = kwargs["segmentation_params"][k]

    kwargs["segmentation_params"] = default_seg_params
    return pimms.plan(**segmentation_tasks)



# This updates usage/methods.rst to the latest API methods
# Developers can run this after modifying any arguments the user can see

from AFQ.api.utils import methods_descriptors
from textwrap import dedent

prologue = """
.. _methods_docs:

The pyAFQ API methods
---------------------
After defining your pyAFQ API object, you can ask for the output of
any step of the pipeline. It is common for users to just call export_all
(for example, `myafq.export_all()`). However, if the user only wants the
tractography, the user can instead call `myafq.export("streamlines")`. Here
is a list of all of pyAFQ's possible outputs:

"""


def setup(app):
    method_descriptions = ""
    for output, desc in methods_descriptors.items():
        desc = desc.replace("\n", " ").replace("\t", "").replace("    ", "")
        method_descriptions = method_descriptions + dedent(f"""

        {output}:
            {desc}
        """)

    with open('./source/reference/methods.rst', 'w') as ff:
        ff.write(prologue + method_descriptions)



import numpy as np
import numpy.testing as npt

import AFQ.data.fetch as afd
import nibabel as nib


def test_read_roi():
    aff1 = np.eye(4)
    template = nib.Nifti1Image(np.ones((10, 10, 10)), aff1)
    aff2 = aff1[:]
    aff2[0, 0] = -1
    roi = nib.Nifti1Image(np.zeros((10, 10, 10)), aff2)
    img = afd.read_resample_roi(roi, resample_to=template)
    npt.assert_equal(img.affine, template.affine)



import numpy as np
import numpy.testing as npt
import nibabel as nib
from AFQ.utils import streamlines as aus
import dipy.tracking.streamline as dts
from dipy.io.stateful_tractogram import StatefulTractogram, Space


def test_SegmentedSFT():

    affine = np.array([[2., 0., 0., -80.],
                       [0., 2., 0., -120.],
                       [0., 0., 2., -60.],
                       [0., 0., 0., 1.]])
    img = nib.Nifti1Image(np.ones((10, 10, 10, 30)), affine)

    bundles = {'ba': StatefulTractogram([np.array([[0, 0, 0],
                                                   [0, 0, 0.5],
                                                   [0, 0, 1],
                                                   [0, 0, 1.5]]),
                                         np.array([[0, 0, 0],
                                                   [0, 0.5, 0.5],
                                                   [0, 1, 1]])],
                                         img,
                                         Space.VOX),
               'bb': StatefulTractogram([np.array([[0, 0, 0],
                                                   [0, 0, 0.5],
                                                   [0, 0, 2],
                                                   [0, 0, 2.5]]),
                                         np.array([[0, 0, 0],
                                                   [0, 0.5, 0.5],
                                                   [0, 2, 2]])],
                                         img,
                                         Space.VOX)}

    seg_sft = aus.SegmentedSFT(bundles, Space.VOX)
    for k1 in bundles.keys():
        for sl1, sl2 in zip(
                bundles[k1].streamlines,
                seg_sft.get_bundle(k1).streamlines):
            npt.assert_equal(sl1, sl2)


def test_split_streamline():
    streamlines = dts.Streamlines([np.array([[1.,2.,3.],
                                    [4.,5.,6.]]),
                                   np.array([[7.,8.,9.],
                                    [10.,11.,12.],
                                    [13., 14., 15.]])])
    assert streamlines == streamlines
    sl_to_split = 1
    split_idx = 1
    new_streamlines = aus.split_streamline(streamlines, sl_to_split, split_idx)
    test_streamlines = dts.Streamlines([np.array([[1., 2., 3.],
                                                  [4., 5., 6.]]),
                                        np.array([[7., 8., 9.]]),
                                        np.array([[10., 11., 12.],
                                                  [13., 14., 15.]])])

    # Test equality of the underlying dict items:
    for k in new_streamlines.__dict__.keys():
        if isinstance(new_streamlines.__dict__[k], np.ndarray):
            npt.assert_array_equal(
                new_streamlines.__dict__[k],
                test_streamlines.__dict__[k]
                )
        else:
            assert new_streamlines.__dict__[k] == test_streamlines.__dict__[k]



import nibabel as nib
import os.path as op
from time import time
import logging
from tqdm import tqdm
import numpy as np
import tempfile
import math

from PIL import Image, ImageDraw, ImageFont

from AFQ.definitions.mapping import SlrMap
from AFQ.api.utils import (
    check_attribute, AFQclass_doc,
    export_all_helper, valid_exports_string)

from AFQ.tasks.data import get_data_plan
from AFQ.tasks.mapping import get_mapping_plan
from AFQ.tasks.tractography import get_tractography_plan
from AFQ.tasks.segmentation import get_segmentation_plan
from AFQ.tasks.viz import get_viz_plan
from AFQ.tasks.utils import get_base_fname
from AFQ.utils.path import apply_cmd_to_afq_derivs
from AFQ.viz.utils import BEST_BUNDLE_ORIENTATIONS, trim, get_eye


__all__ = ["ParticipantAFQ"]


class ParticipantAFQ(object):
    f"""{AFQclass_doc}"""

    def __init__(self,
                 dwi_data_file,
                 bval_file, bvec_file,
                 output_dir,
                 **kwargs):
        """
        Initialize a ParticipantAFQ object.

        Parameters
        ----------
        dwi_data_file : str
            Path to DWI data file.
        bval_file : str
            Path to bval file.
        bvec_file : str
            Path to bvec file.
        output_dir : str
            Path to output directory.
        kwargs : additional optional parameters
            You can set additional parameters for any step
            of the process. See :ref:`usage/kwargs` for more details.

        Examples
        --------
        api.ParticipantAFQ(
            dwi_data_file, bval_file, bvec_file, output_dir,
            csd_sh_order=4)
        api.ParticipantAFQ(
            dwi_data_file, bval_file, bvec_file, output_dir,
            reg_template_spec="mni_t2", reg_subject_spec="b0")

        Notes
        -----
        In tracking_params, parameters with the suffix mask which are also
        an image from AFQ.definitions.image will be handled automatically by
        the api.
        """
        if not isinstance(output_dir, str):
            raise TypeError(
                "output_dir must be a str")
        if not isinstance(dwi_data_file, str):
            raise TypeError(
                "dwi_data_file must be a str")
        if not isinstance(bval_file, str):
            raise TypeError(
                "bval_file must be a str")
        if not isinstance(bvec_file, str):
            raise TypeError(
                "bvec_file must be a str")
        if not op.exists(output_dir):
            raise ValueError(
                f"output_dir does not exist: {output_dir}")
        if "tractography_params" in kwargs:
            raise ValueError((
                "unrecognized parameter tractography_params, "
                "did you mean tracking_params ?"))

        self.logger = logging.getLogger('AFQ')

        self.kwargs = dict(
            dwi_data_file=dwi_data_file,
            bval_file=bval_file,
            bvec_file=bvec_file,
            output_dir=output_dir,
            base_fname=get_base_fname(output_dir, dwi_data_file),
            **kwargs)
        self.make_workflow()

    def make_workflow(self):
        # construct pimms plans
        if "mapping_definition" in self.kwargs and isinstance(
                self.kwargs["mapping_definition"], SlrMap):
            plans = {  # if using SLR map, do tractography first
                "data": get_data_plan(self.kwargs),
                "tractography": get_tractography_plan(
                    self.kwargs
                ),
                "mapping": get_mapping_plan(
                    self.kwargs,
                    use_sls=True
                ),
                "segmentation": get_segmentation_plan(self.kwargs),
                "viz": get_viz_plan(self.kwargs)}
        else:
            plans = {  # Otherwise, do mapping first
                "data": get_data_plan(self.kwargs),
                "mapping": get_mapping_plan(self.kwargs),
                "tractography": get_tractography_plan(
                    self.kwargs
                ),
                "segmentation": get_segmentation_plan(self.kwargs),
                "viz": get_viz_plan(self.kwargs)}

        # chain together a complete plan from individual plans
        previous_data = {}
        for name, plan in plans.items():
            previous_data[f"{name}_imap"] = plan(
                **self.kwargs,
                **previous_data)

        self.wf_dict =\
            previous_data[f"{name}_imap"]

    def export(self, attr_name="help"):
        """
        Export a specific output. To print a list of available outputs,
        call export without arguments.

        Parameters
        ----------
        attr_name : str
            Name of the output to export. Default: "help"

        Returns
        -------
        output : any
            The specific output, or None if called without arguments.
        """
        section = check_attribute(attr_name)

        if section is None:
            return self.wf_dict[attr_name]
        return self.wf_dict[section][attr_name]

    def export_up_to(self, attr_name="help"):
        f"""
        Export all derivatives necessary for a specific output.
        To print a list of available outputs,
        call export_up_to without arguments.
        {valid_exports_string}

        Parameters
        ----------
        attr_name : str
            Name of the output to export up to. Default: "help"
        """
        section = check_attribute(attr_name)
        wf_dict = self.wf_dict
        if section is not None:
            wf_dict = wf_dict[section]
        for dependent in wf_dict.plan.dependencies[attr_name]:
            self.export(dependent)

    def export_all(self, viz=True, xforms=True,
                   indiv=True):
        f""" Exports all the possible outputs
        {valid_exports_string}

        Parameters
        ----------
        viz : bool
            Whether to output visualizations. This includes tract profile
            plots, a figure containing all bundles, and, if using the AFQ
            segmentation algorithm, individual bundle figures.
            Default: True
        xforms : bool
            Whether to output the reg_template image in subject space and,
            depending on if it is possible based on the mapping used, to
            output the b0 in template space.
            Default: True
        indiv : bool
            Whether to output individual bundles in their own files, in
            addition to the one file containing all bundles. If using
            the AFQ segmentation algorithm, individual ROIs are also
            output.
            Default: True
        """
        start_time = time()
        export_all_helper(self, xforms, indiv, viz)
        self.logger.info(
            f"Time taken for export all: {time() - start_time}")

    def participant_montage(self, images_per_row=2):
        """
        Generate montage of all bundles for a given subject.

        Parameters
        ----------
        images_per_row : int
            Number of bundle images per row in output file.
            Default: 2

        Returns
        -------
        filename of montage images
        """
        tdir = tempfile.gettempdir()

        all_fnames = []
        bundle_dict = self.export("bundle_dict")
        self.logger.info("Generating Montage...")
        viz_backend = self.export("viz_backend")
        best_scalar = self.export(self.export("best_scalar"))
        size = (images_per_row, math.ceil(len(bundle_dict) / images_per_row))
        for ii, bundle_name in enumerate(tqdm(bundle_dict)):
            flip_axes = [False, False, False]
            for i in range(3):
                flip_axes[i] = (self.export("dwi_affine")[i, i] < 0)

            figure = viz_backend.visualize_volume(
                best_scalar,
                flip_axes=flip_axes,
                interact=False,
                inline=False)
            figure = viz_backend.visualize_bundles(
                self.export("bundles"),
                shade_by_volume=best_scalar,
                color_by_direction=True,
                flip_axes=flip_axes,
                bundle=bundle_name,
                figure=figure,
                interact=False,
                inline=False)

            view, direc = BEST_BUNDLE_ORIENTATIONS.get(
                bundle_name, ("Axial", "Top"))
            eye = get_eye(view, direc)

            this_fname = tdir + f"/t{ii}.png"
            if "plotly" in viz_backend.backend:
                figure.update_layout(
                    scene_camera=dict(
                        projection=dict(type="orthographic"),
                        up={"x": 0, "y": 0, "z": 1},
                        eye=eye,
                        center=dict(x=0, y=0, z=0)),
                    showlegend=False)
                figure.write_image(this_fname, scale=4)

                # temporary fix for memory leak
                import plotly.io as pio
                pio.kaleido.scope._shutdown_kaleido()
            else:
                from dipy.viz import window
                direc = np.fromiter(eye.values(), dtype=int)
                data_shape = np.asarray(
                    nib.load(self.export("b0")).get_fdata().shape)
                figure.set_camera(
                    position=direc * data_shape,
                    focal_point=data_shape // 2,
                    view_up=(0, 0, 1))
                figure.zoom(0.5)
                window.snapshot(figure, fname=this_fname, size=(600, 600))

        def _save_file(curr_img):
            save_path = op.abspath(op.join(
                self.kwargs["output_dir"],
                "bundle_montage.png"))
            curr_img.save(save_path)
            all_fnames.append(save_path)

        this_img_trimmed = {}
        max_height = 0
        max_width = 0
        for ii, bundle_name in enumerate(bundle_dict):
            this_img = Image.open(tdir + f"/t{ii}.png")
            try:
                this_img_trimmed[ii] = trim(this_img)
            except IndexError:  # this_img is a picture of nothing
                this_img_trimmed[ii] = this_img

            text_sz = 70
            width, height = this_img_trimmed[ii].size
            height = height + text_sz
            result = Image.new(
                this_img_trimmed[ii].mode, (width, height),
                color=(255, 255, 255))
            result.paste(this_img_trimmed[ii], (0, text_sz))
            this_img_trimmed[ii] = result

            draw = ImageDraw.Draw(this_img_trimmed[ii])
            draw.text(
                (0, 0), bundle_name, (0, 0, 0),
                font=ImageFont.truetype(
                    "Arial", text_sz))

            if this_img_trimmed[ii].size[0] > max_width:
                max_width = this_img_trimmed[ii].size[0]
            if this_img_trimmed[ii].size[1] > max_height:
                max_height = this_img_trimmed[ii].size[1]

        curr_img = Image.new(
            'RGB',
            (max_width * size[0], max_height * size[1]),
            color="white")

        for ii in range(len(bundle_dict)):
            x_pos = ii % size[0]
            _ii = ii // size[0]
            y_pos = _ii % size[1]
            _ii = _ii // size[1]
            this_img = this_img_trimmed[ii].resize((max_width, max_height))
            curr_img.paste(
                this_img,
                (x_pos * max_width, y_pos * max_height))

        _save_file(curr_img)
        return all_fnames

    def cmd_outputs(self, cmd="rm", dependent_on=None, exceptions=[],
                    suffix=""):
        """
        Perform some command some or all outputs of pyafq.
        This is useful if you change a parameter and need
        to recalculate derivatives that depend on it.
        Some examples: cp, mv, rm .
        -r will be automtically added when necessary.

        Parameters
        ----------
        cmd : str
            Command to run on outputs. Default: 'rm'
        dependent_on : str or None
            Which derivatives to perform command on .
            If None, perform on all.
            If "track", perform on all derivatives that depend on the
            tractography.
            If "recog", perform on all derivatives that depend on the
            bundle recognition.
            If "prof", perform on all derivatives that depend on the
            bundle profiling.
            Default: None
        exceptions : list of str
            Name outputs that the command should not be applied to.
            Default: []
        suffix : str
            Parts of command that are used after the filename.
            Default: ""
        """
        exception_file_names = []
        for exception in exceptions:
            file_name = self.export(exception)
            if isinstance(file_name, str):
                exception_file_names.append(file_name)
            else:
                self.logger.warning((
                    f"The exception '{exception}' does not correspond"
                    " to a filename and will be ignored."))

        apply_cmd_to_afq_derivs(
            self.kwargs["output_dir"],
            self.export("base_fname"),
            cmd=cmd,
            exception_file_names=exception_file_names,
            suffix=suffix,
            dependent_on=dependent_on
        )

        # do not assume previous calculations are still valid
        # after file operations
        self.make_workflow()

    clobber = cmd_outputs  # alias for default of cmd_outputs



"""
=================================================
Making videos the different stages of tractometry
=================================================

Two-dimensional figures of anatomical data are somewhat limited, because of the
complex three-dimensional configuration of the brain. Therefored, dynamic
videos of anatomical data are useful for exploring the data, as well as for
creating dynamic presentations of the results. This example visualizes various
stages of tractometry, from preprocessed diffusion data to the final tract
profiles. We will use the `Fury <https://fury.gl/>`_ software library to
visualize individual frames of the results of each stage, and then create
videos of each stage of the process using the Python Image Library (PIL, also
known as pillow).

"""

##############################################################################
# Imports
# -------
#


import os
import os.path as op
import nibabel as nib
import numpy as np
import tempfile

from dipy.io.streamline import load_trk
from dipy.tracking.streamline import (transform_streamlines,
                                      set_number_of_points)
from dipy.core.gradients import gradient_table
from dipy.align import resample

from fury import actor, window
from fury.actor import colormap_lookup_table
from fury.colormap import create_colormap
from matplotlib.cm import tab20

import AFQ.data.fetch as afd
from AFQ.viz.utils import gen_color_dict

from PIL import Image

##############################################################################
# Define a function that makes videos
# -----------------------------------
# The PIL library has a function that can be used to create animated GIFs from
# a series of images. We will use this function to create videos.
#
# .. note::
#  This function is not part of the AFQ library, but is included here for
#  convenience. It is not necessary to understand this function in order to
#  understand the rest of the example. If you are interested in learning more
#  about this function, you can read the PIL documentation. The function is
#  based on the `PIL.Image.save <https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.save>`_
#  function.


def make_video(frames, out):
    """
    Make a video from a series of frames.

    Parameters
    ----------
    frames : list of str
        A list of file names of the frames to be included in the video.

    out : str
        The name of the output file. Format is determined by the file
        extension.
    """
    video = []
    for nn in frames:
        frame = Image.open(nn)
        video.append(frame)

    # Save the frames as an animated GIF
    video[0].save(
        out,
        save_all=True,
        append_images=video[1:],
        duration=300,
        loop=1)


tmp = tempfile.mkdtemp()
n_frames = 72

###############################################################################
# Get data from HBN POD2
# ----------------------------
# We get the same data that is used in the visualization tutorials.

afd.fetch_hbn_preproc(["NDARAA948VFH"])
study_path = afd.fetch_hbn_afq(["NDARAA948VFH"])[1]

#############################################################################
# Visualize the processed dMRI data
# ---------------------------------
# The HBN POD2 dataset was processed using the ``qsiprep`` pipeline. The
# results from this processing are stored within a sub-folder of the
# derivatives folder wthin the study folder.
# Here, we will start by visualizing the diffusion data. We read in the
# diffusion data, as well as the gradient table, using the `nibabel` library.
# We then extract the b0, b1000, and b2000 volumes from the diffusion data.
# We will use the `actor.slicer` function from `fury` to visualize these. This
# function takes a 3D volume as input and returns a `slicer` actor, which can
# then be added to a `window.Scene` object. We create a helper function that
# will create a slicer actor for a given volume and a given slice along the x,
# y, or z dimension. We then call this function three times, once for each of
# the b0, b1000, and b2000 volumes, and add the resulting slicer actors to a
# scene. We set the camera on the scene to a view that we like, and then we
# record the scene into png files and subsequently gif animations. We do this
# for each of the three volumes.

deriv_path = op.join(
    study_path, "derivatives")

qsiprep_path = op.join(
    deriv_path,
    'qsiprep',
    'sub-NDARAA948VFH',
    'ses-HBNsiteRU')

dmri_img = nib.load(op.join(
    qsiprep_path,
    'dwi',
    'sub-NDARAA948VFH_ses-HBNsiteRU_acq-64dir_space-T1w_desc-preproc_dwi.nii.gz'))

gtab = gradient_table(*[op.join(
    qsiprep_path,
    'dwi',
    f'sub-NDARAA948VFH_ses-HBNsiteRU_acq-64dir_space-T1w_desc-preproc_dwi.{ext}') for ext in ['bval', 'bvec']])


dmri_data = dmri_img.get_fdata()

dmri_b0 = dmri_data[..., 0]
dmri_b1000 = dmri_data[..., 1]
dmri_b2000 = dmri_data[..., 65]


def slice_volume(data, x=None, y=None, z=None):
    slicer_actors = []
    slicer_actor_z = actor.slicer(data)
    if z is not None:
        slicer_actor_z.display_extent(
            0, data.shape[0] - 1,
            0, data.shape[1] - 1,
            z, z)
        slicer_actors.append(slicer_actor_z)
    if y is not None:
        slicer_actor_y = slicer_actor_z.copy()
        slicer_actor_y.display_extent(
            0, data.shape[0] - 1,
            y, y,
            0, data.shape[2] - 1)
        slicer_actors.append(slicer_actor_y)
    if x is not None:
        slicer_actor_x = slicer_actor_z.copy()
        slicer_actor_x.display_extent(
            x, x,
            0, data.shape[1] - 1,
            0, data.shape[2] - 1)
        slicer_actors.append(slicer_actor_x)

    return slicer_actors


slicers_b0 = slice_volume(
    dmri_b0,
    x=dmri_b0.shape[0] // 2,
    y=dmri_b0.shape[1] // 2,
    z=dmri_b0.shape[-1] // 3)
slicers_b1000 = slice_volume(
    dmri_b1000,
    x=dmri_b0.shape[0] // 2,
    y=dmri_b0.shape[1] // 2,
    z=dmri_b0.shape[-1] // 3)
slicers_b2000 = slice_volume(
    dmri_b2000,
    x=dmri_b0.shape[0] // 2,
    y=dmri_b0.shape[1] // 2,
    z=dmri_b0.shape[-1] // 3)

for bval, slicers in zip([0, 1000, 2000],
                         [slicers_b0, slicers_b1000, slicers_b2000]):
    scene = window.Scene()
    for slicer in slicers:
        scene.add(slicer)
    scene.set_camera(position=(721.34, 393.48, 97.03),
                     focal_point=(96.00, 114.00, 96.00),
                     view_up=(-0.01, 0.02, 1.00))

    scene.background((1, 1, 1))
    window.record(scene, out_path=f'{tmp}/b{bval}',
                  size=(2400, 2400),
                  n_frames=n_frames, path_numbering=True)

    make_video(
        [f'{tmp}/b{bval}{ii:06d}.png' for ii in range(n_frames)], f'b{bval}.gif')
#############################################################################
# Visualizing whole-brain tractography
# ------------------------------------
# One of the first steps of the pyAFQ pipeline is to generate whole-brain
# tractography. We will visualize the results of this step. We start by reading
# in the FA image, which is used as a reference for the tractography. We then
# load the whole brain tractography, and transform the coordinates of the
# streamlines into the coordinate frame of the T1-weighted data.

afq_path = op.join(
    deriv_path,
    'afq',
    'sub-NDARAA948VFH',
    'ses-HBNsiteRU')

fa_img = nib.load(op.join(afq_path,
                          'sub-NDARAA948VFH_ses-HBNsiteRU_acq-64dir_space-T1w_desc-preproc_dwi_model-DKI_FA.nii.gz'))


sft_whole_brain = load_trk(op.join(afq_path,
                                   'sub-NDARAA948VFH_ses-HBNsiteRU_acq-64dir_space-T1w_desc-preproc_dwi_space-RASMM_model-CSD_desc-prob-afq_tractography.trk'), fa_img)


t1w_img = nib.load(op.join(deriv_path,
                           'qsiprep/sub-NDARAA948VFH/anat/sub-NDARAA948VFH_desc-preproc_T1w.nii.gz'))
t1w = t1w_img.get_fdata()
sft_whole_brain.to_rasmm()
whole_brain_t1w = transform_streamlines(
    sft_whole_brain.streamlines,
    np.linalg.inv(t1w_img.affine))

#############################################################################
# Visualize the streamlines
# -------------------------
# The streamlines are visualized in the context of the T1-weighted data.
#


def lines_as_tubes(sl, line_width, **kwargs):
    line_actor = actor.line(sl, **kwargs)
    line_actor.GetProperty().SetRenderLinesAsTubes(1)
    line_actor.GetProperty().SetLineWidth(line_width)
    return line_actor


whole_brain_actor = lines_as_tubes(whole_brain_t1w, 2)
slicers = slice_volume(t1w, y=t1w.shape[1] // 2 - 5, z=t1w.shape[-1] // 3)

scene = window.Scene()

scene.add(whole_brain_actor)
for slicer in slicers:
    scene.add(slicer)

scene.set_camera(position=(721.34, 393.48, 97.03),
                 focal_point=(96.00, 114.00, 96.00),
                 view_up=(-0.01, 0.02, 1.00))

scene.background((1, 1, 1))
window.record(scene, out_path=f'{tmp}/whole_brain', size=(2400, 2400),
              n_frames=n_frames, path_numbering=True)

make_video([f"{tmp}/whole_brain{ii:06d}.png" for ii in range(n_frames)],
           "whole_brain.gif")

#############################################################################
# Whole brain with waypoints
# --------------------------------------
# We can also generate a gif video with the whole brain tractography and the
# waypoints that are used to define the bundles. We will use the same scene as
# before, but we will add the waypoints as contours to the scene.

scene.clear()
whole_brain_actor = lines_as_tubes(whole_brain_t1w, 2)

scene.add(whole_brain_actor)
for slicer in slicers:
    scene.add(slicer)

scene.background((1, 1, 1))

waypoint1 = nib.load(
    op.join(
        afq_path,
        "ROIs", "sub-NDARAA948VFH_ses-HBNsiteRU_acq-64dir_space-T1w_desc-preproc_dwi_desc-ROI-ARC_L-1-include.nii.gz"))

waypoint2 = nib.load(
    op.join(
        afq_path,
        "ROIs", "sub-NDARAA948VFH_ses-HBNsiteRU_acq-64dir_space-T1w_desc-preproc_dwi_desc-ROI-ARC_L-2-include.nii.gz"))

waypoint1_xform = resample(waypoint1, t1w_img)
waypoint2_xform = resample(waypoint2, t1w_img)
waypoint1_data = waypoint1_xform.get_fdata() > 0
waypoint2_data = waypoint2_xform.get_fdata() > 0

surface_color = tab20.colors[0]

waypoint1_actor = actor.contour_from_roi(waypoint1_data,
                                         color=surface_color,
                                         opacity=0.5)

waypoint2_actor = actor.contour_from_roi(waypoint2_data,
                                         color=surface_color,
                                         opacity=0.5)

scene.add(waypoint1_actor)
scene.add(waypoint2_actor)

window.record(scene, out_path=f'{tmp}/whole_brain_with_waypoints', size=(2400, 2400),
              n_frames=n_frames, path_numbering=True)

make_video([f"{tmp}/whole_brain_with_waypoints{ii:06d}.png" for ii in range(n_frames)],
           "whole_brain_with_waypoints.gif")

bundle_path = op.join(afq_path,
                      'bundles')


#############################################################################
# Visualize the arcuate bundle
# ----------------------------
# Now visualize only the arcuate bundle that is selected with these waypoints.
#

fa_img = nib.load(op.join(afq_path,
                          'sub-NDARAA948VFH_ses-HBNsiteRU_acq-64dir_space-T1w_desc-preproc_dwi_model-DKI_FA.nii.gz'))
fa = fa_img.get_fdata()
sft_arc = load_trk(op.join(bundle_path,
                           'sub-NDARAA948VFH_ses-HBNsiteRU_acq-64dir_space-T1w_desc-preproc_dwi_space-RASMM_model-CSD_desc-prob-afq-ARC_L_tractography.trk'), fa_img)

sft_arc.to_rasmm()
arc_t1w = transform_streamlines(sft_arc.streamlines,
                                np.linalg.inv(t1w_img.affine))


bundles = [
    "ARC_R",
    "ATR_R",
    "CST_R",
    "IFO_R",
    "ILF_R",
    "SLF_R",
    "UNC_R",
    "CGC_R",
    "Orbital", "AntFrontal", "SupFrontal", "Motor",
    "SupParietal", "PostParietal", "Temporal", "Occipital",
    "CGC_L",
    "UNC_L",
    "SLF_L",
    "ILF_L",
    "IFO_L",
    "CST_L",
    "ATR_L",
    "ARC_L",
]

color_dict = gen_color_dict(bundles)

arc_actor = lines_as_tubes(arc_t1w, 8, colors=color_dict['ARC_L'])
scene.clear()

scene.add(arc_actor)
for slicer in slicers:
    scene.add(slicer)

scene.add(waypoint1_actor)
scene.add(waypoint2_actor)

window.record(scene, out_path=f'{tmp}/arc1', size=(2400, 2400),
              n_frames=n_frames, path_numbering=True)

make_video([f"{tmp}/arc1{ii:06d}.png" for ii in range(n_frames)], "arc1.gif")

#############################################################################
# Clean bundle
# ------------
# The next step in processing would be to clean the bundle by removing
# streamlines that are outliers. We will visualize the cleaned bundle.

scene.clear()

scene.add(arc_actor)
for slicer in slicers:
    scene.add(slicer)

window.record(scene, out_path=f'{tmp}/arc2', size=(2400, 2400),
              n_frames=n_frames, path_numbering=True)

make_video([f"{tmp}/arc2{ii:06d}.png" for ii in range(n_frames)], "arc2.gif")

clean_bundles_path = op.join(afq_path,
                             'clean_bundles')

sft_arc = load_trk(op.join(clean_bundles_path,
                           'sub-NDARAA948VFH_ses-HBNsiteRU_acq-64dir_space-T1w_desc-preproc_dwi_space-RASMM_model-CSD_desc-prob-afq-ARC_L_tractography.trk'), fa_img)

sft_arc.to_rasmm()
arc_t1w = transform_streamlines(sft_arc.streamlines,
                                np.linalg.inv(t1w_img.affine))


arc_actor = lines_as_tubes(arc_t1w, 8, colors=tab20.colors[18])
scene.clear()

scene.add(arc_actor)
for slicer in slicers:
    scene.add(slicer)

window.record(scene, out_path=f'{tmp}/arc3', size=(2400, 2400),
              n_frames=n_frames, path_numbering=True)

make_video([f"{tmp}/arc3{ii:06d}.png" for ii in range(n_frames)], "arc3.gif")

#############################################################################
# Show the values of tissue properties along the bundle
# ------------------------------------------------------
# We can also visualize the values of tissue properties along the bundle. Here
# we will visualize the fractional anisotropy (FA) along the arcuate bundle.
# This is done by using a colormap to color the streamlines according to the
# values of the tissue property, with `fury.colormap.create_colormap`.

lut_args = dict(scale_range=(0, 1),
                hue_range=(1, 0),
                saturation_range=(0, 1),
                value_range=(0, 1))

arc_actor = lines_as_tubes(arc_t1w, 8,
                           colors=resample(fa_img, t1w_img).get_fdata(),
                           lookup_colormap=colormap_lookup_table(**lut_args))
scene.clear()

scene.add(arc_actor)
for slicer in slicers:
    scene.add(slicer)

window.record(scene, out_path=f'{tmp}/arc4', size=(2400, 2400),
              n_frames=n_frames, path_numbering=True)

make_video([f"{tmp}/arc4{ii:06d}.png" for ii in range(n_frames)], "arc4.gif")

#############################################################################
# Core of the bundle and tract profile
# -------------------------------------
# Finally, we can visualize the core of the bundle and the tract profile. The
# core of the bundle is the median of the streamlines, and the tract profile is
# the values of the tissue property along the core of the bundle.

core_arc = np.median(np.asarray(set_number_of_points(arc_t1w, 20)), axis=0)

from dipy.stats.analysis import afq_profile
sft_arc.to_vox()
arc_profile = afq_profile(fa, sft_arc.streamlines, affine=np.eye(4),
                          n_points=20)

core_arc_actor = lines_as_tubes(
    [core_arc],
    40,
    colors=create_colormap(arc_profile, 'viridis')
)

arc_actor = lines_as_tubes(arc_t1w, 1,
                           colors=resample(fa_img, t1w_img).get_fdata(),
                           lookup_colormap=colormap_lookup_table(**lut_args))

scene.clear()

for slicer in slicers:
    scene.add(slicer)
scene.add(arc_actor)
scene.add(core_arc_actor)

window.record(scene, out_path=f'{tmp}/arc5', size=(2400, 2400),
              n_frames=n_frames, path_numbering=True)

make_video([f"{tmp}/arc5{ii:06d}.png" for ii in range(n_frames)], "arc5.gif")

#############################################################################
# Core of all bundles and their tract profiles
# --------------------------------------------
# Same as before, but for all bundles.

scene.clear()

for slicer in slicers:
    scene.add(slicer)

for bundle in bundles:
    sft = load_trk(op.join(clean_bundles_path,
                           f'sub-NDARAA948VFH_ses-HBNsiteRU_acq-64dir_space-T1w_desc-preproc_dwi_space-RASMM_model-CSD_desc-prob-afq-{bundle}_tractography.trk'), fa_img)

    sft.to_rasmm()
    bundle_t1w = transform_streamlines(sft.streamlines,
                                       np.linalg.inv(t1w_img.affine))

    bundle_actor = lines_as_tubes(bundle_t1w, 8, colors=color_dict[bundle])
    scene.add(bundle_actor)

window.record(scene, out_path=f'{tmp}/all_bundles', size=(2400, 2400),
              n_frames=n_frames, path_numbering=True)

make_video(
    [f"{tmp}/all_bundles{ii:06d}.png" for ii in range(n_frames)], "all_bundles.gif")


scene.clear()

for slicer in slicers:
    scene.add(slicer)

tract_profiles = []
for bundle in bundles:
    sft = load_trk(op.join(clean_bundles_path,
                           f'sub-NDARAA948VFH_ses-HBNsiteRU_acq-64dir_space-T1w_desc-preproc_dwi_space-RASMM_model-CSD_desc-prob-afq-{bundle}_tractography.trk'), fa_img)
    sft.to_rasmm()
    bundle_t1w = transform_streamlines(sft.streamlines,
                                       np.linalg.inv(t1w_img.affine))

    core_bundle = np.median(np.asarray(
        set_number_of_points(bundle_t1w, 20)), axis=0)
    sft.to_vox()
    tract_profiles.append(
        afq_profile(fa, sft.streamlines, affine=np.eye(4),
                    n_points=20))

    core_actor = lines_as_tubes(
        [core_bundle],
        40,
        colors=create_colormap(tract_profiles[-1], 'viridis')
    )

    scene.add(core_actor)

window.record(scene,
              out_path=f'{tmp}/all_tract_profiles',
              size=(2400, 2400),
              n_frames=n_frames,
              path_numbering=True)

make_video([f"{tmp}/all_tract_profiles{ii:06d}.png" for ii in range(n_frames)],
           "all_tract_profiles.gif")

#############################################################################
# Tract profiles as a table
# -------------------------
# Finally, we can visualize the tract profiles as a table. This is done by
# plotting the tract profiles for each bundle as a line plot, with the x-axis
# representing the position along the bundle, and the y-axis representing the
# value of the tissue property. We will use the `matplotlib` library to create
# this plot.

import matplotlib.pyplot as plt
fig, ax = plt.subplots()
for ii, bundle in enumerate(bundles):
    ax.plot(np.arange(ii * 20, (ii + 1) * 20),
            tract_profiles[ii],
            color=color_dict[bundle],
            linewidth=3)
ax.set_xticks(np.arange(0, 20 * len(bundles), 20))
ax.set_xticklabels(bundles, rotation=45, ha='right')
fig.set_size_inches(10, 5)
plt.subplots_adjust(bottom=0.2)
fig.savefig('tract_profiles_as_table.png')



import os.path as op

import numpy as np
import numpy.testing as npt

import nibabel as nib
import nibabel.tmpdirs as nbtmp

import dipy.data as dpd

from AFQ.registration import (write_mapping,
                              read_mapping, syn_register_dwi,
                              slr_registration)

import AFQ.data.fetch as afd

from dipy.io.streamline import load_tractogram
from dipy.align.imwarp import DiffeomorphicMap


MNI_T2 = afd.read_mni_template()
hardi_img, gtab = dpd.read_stanford_hardi()
MNI_T2_data = MNI_T2.get_fdata()
MNI_T2_affine = MNI_T2.affine
hardi_data = hardi_img.get_fdata()
hardi_affine = hardi_img.affine
b0 = hardi_data[..., gtab.b0s_mask]
mean_b0 = np.mean(b0, -1)

# We select some arbitrary chunk of data so this goes quicker:
subset_b0 = mean_b0[40:50, 40:50, 40:50]
subset_dwi_data = nib.Nifti1Image(hardi_data[40:50, 40:50, 40:50],
                                  hardi_affine)
subset_t2 = MNI_T2_data[40:60, 40:60, 40:60]
subset_b0_img = nib.Nifti1Image(subset_b0, hardi_affine)
subset_t2_img = nib.Nifti1Image(subset_t2, MNI_T2_affine)


def test_slr_registration():
    # have to import subject sls
    file_dict = afd.read_stanford_hardi_tractography()
    streamlines = file_dict['tractography_subsampled.trk']

    # have to import sls atlas
    afd.fetch_hcp_atlas_16_bundles()
    atlas_fname = op.join(
        afd.afq_home,
        'hcp_atlas_16_bundles',
        'Atlas_in_MNI_Space_16_bundles',
        'whole_brain',
        'whole_brain_MNI.trk')
    hcp_atlas = load_tractogram(
        atlas_fname,
        'same', bbox_valid_check=False)

    with nbtmp.InTemporaryDirectory() as tmpdir:
        mapping = slr_registration(streamlines,
                                   hcp_atlas.streamlines,
                                   moving_affine=subset_b0_img.affine,
                                   static_affine=subset_t2_img.affine,
                                   moving_shape=subset_b0_img.shape,
                                   static_shape=subset_t2_img.shape,
                                   progressive=False,
                                   greater_than=10,
                                   rm_small_clusters=1,
                                   rng=np.random.RandomState(seed=8))
        warped_moving = mapping.transform(subset_b0)

        npt.assert_equal(warped_moving.shape, subset_t2.shape)
        mapping_fname = op.join(tmpdir, 'mapping.npy')
        write_mapping(mapping, mapping_fname)
        file_mapping = read_mapping(mapping_fname,
                                    subset_b0_img,
                                    subset_t2_img)

        # Test that it has the same effect on the data:
        warped_from_file = file_mapping.transform(subset_b0)
        npt.assert_equal(warped_from_file, warped_moving)

        # Test that it is, attribute by attribute, identical:
        for k in mapping.__dict__:
            assert (np.all(mapping.__getattribute__(k) ==
                           file_mapping.__getattribute__(k)))


def test_syn_register_dwi():
    warped_b0, mapping = syn_register_dwi(subset_dwi_data, gtab,
                                          template=subset_t2_img,
                                          radius=1)
    npt.assert_equal(isinstance(mapping, DiffeomorphicMap), True)
    npt.assert_equal(warped_b0.shape, subset_t2_img.shape)



"""
=============================================
BabyAFQ : tractometry for infant dMRI data
=============================================

The following is an example of tractometry for infant bundles. This example and
resulting pyAFQ support for pediatric bundles was inspired by and largely due
to the work of Grotheer et al. [Grotheer2022]_, as implemented in
[Grotheer2023]_.

.. note::
    Because it is time and disk-space consuming, this example
    is not run when the pyAFQ documentation is built. To run this example
    yourself, you can download the contents of this file as an
    executable `.py` file or as a Jupyter notebook from the links at the bottom
    of the page.

"""
import os.path as op
import plotly
import wget
import zipfile

from AFQ.api.group import GroupAFQ
import AFQ.api.bundle_dict as abd
import AFQ.data.fetch as afd

##########################################################################
# Baby dMRI data
# -------------------------
#
# Infant MRI data are quite different from data acquired in grownup
# participants, and even from children that are just a few years older.
# First, there is the rather obvious difference in size. Baby brains are
# approximately 25% the size of grownup brains at birth. But there are also
# often less known differences in brain tissue properties. For example, the
# myelin content of white matter is much lower in infants than in grownups.
# This is important because the diffusion signal that we measure with dMRI is
# sensitive to the myelin content, and it means that the dMRI signal differs
# quite a bit in newborn infants. For the purpose of delineating the major
# white matter pathways, it is also important to know that their shape,
# location and curvature is different in infants than in grownups. For example,
# the arcuate fasciculus is much more curved in infants than in grownups.
# Because of this, we use a different set of templates for infant brains than
# for grownup brains. These templates were created and validated by
# Mareike Grotheer and colleagues in [Grotheer2022]_. They are available for
# download as part of the pyAFQ software, as we will show below.
#
# In this example, we will demonstrate the use of pyAFQ on data from one
# infant. The data, provided by Kalanit Grill Spector's
# `Stanford Vision and Perception Neuroscience Lab <http://vpnl.stanford.edu/>`,
# and was previously published in [Grotheer2022]_.
# The data is available to download on
# `Figshare <https://figshare.com/articles/dataset/Example_babyAFQ_BIDS_subject/21440739>`.
# You can download it from there and unzip it into ~/AFQ_Data/baby_example/
# (Note that this is 2.69GB of data, so it can take a while to download). Or
# you can download it and unzip it using the following block of code.

data_folder = op.join(op.expanduser('~'), "AFQ_data/")
baby_zip = op.join(data_folder, "baby_example.zip")
if not op.exists(baby_zip):
    print("Downloading processed pediatric data; this could take a while...")
    wget.download(
        "https://figshare.com/ndownloader/files/38053692",
        baby_zip)

with zipfile.ZipFile(baby_zip, 'r') as zip_ref:
    zip_ref.extractall(op.join(data_folder, "baby_example"))

##########################################################################
# Initialize a GroupAFQ object:
# -----------------------------
#
# Now that the data is downloaded and organized in a BIDS-compliant structure,
# we can start running pyAFQ on it. We start by initializing a GroupAFQ object.
# This object manages all of the data transformations and computations
# conducted by the software, based on its initial configuration, which we set
# up below.
#
# A few special things to note here:
#
# 1. The data were preprocessed using the `vistasoft` pipeline, so we set
#    `preproc_pipeline = "vistasoft"`.
# 2. We use the UNC neonatal template, which can be read on a call to the
#    `read_pediatric_templates` function in `AFQ.data.fetch`.
# 3. We use the `baby_bd` to define the bundles that we want to
#    segment. This dictionary is different from the default behavior in that it
#    uses the waypoint ROIs from [Grotheer2022]_.
# 4. In this case, tractography has already been run using
#    `MRTRIX <https://www.mrtrix.org/>`, and is accessed using the
#    `import_tract` key-word argument.

myafq = GroupAFQ(
    bids_path=op.join(op.expanduser('~'),
                      "AFQ_data/baby_example/example_bids_subject"),
    preproc_pipeline="vistasoft",
    reg_template_spec=afd.read_pediatric_templates(
    )["UNCNeo-withCerebellum-for-babyAFQ"],
    reg_subject_spec="b0",
    bundle_info=abd.baby_bd(),
    import_tract={
        "suffix": "tractography", "scope": "mrtrix"},
)

##########################################################################
# Running the pipeline
# --------------------
#
# A call to the `export` function will trigger the pyAFQ pipeline. This will
# run tractography, bundle segmentation, and bundle cleaning. The results will
# be saved in the `~/AFQ_data/baby_example/derivatives/afq` folder. This can
# take a while to run, depending on your computer.
# In this case, we call `export` with the `all_bundles_figure` option. This is
# because visualizations are created after most other parts of the pipeline
# have been run. This means that when this call is done, you should have many
# of the derivative results in the output folder, including the tractography,
# segmentation, and tract profile results, as well as the visualizations.

viz = myafq.export("all_bundles_figure")

##########################################################################
# Viewing the results
# -------------------
# One way to view the results is to open the file named
# `sub-01_ses-01_dwi_space-RASMM_model-probCSD_algo-AFQ_desc-viz_dwi.html`
# in your browser. This is a visualization of the tractography and segmentation
# results for all of the bundles. You can navigate this visualization by
# clicking on the different bundles in the legend on the right side of the
# screen. You can also zoom in and out using the mouse wheel, and rotate the
# view by clicking and dragging with the mouse. You can also view the FA tract
# profiles in a plot on the left side of the page.
#
# If the baby bundles appear dark in the html visualization due to low FA values, you
# can reduce the upper limit of the range in the `sbv_lims_bundles` option when
# building your GroupAFQ object (e.g. `GroupAFQ(..., sbv_lims_bundles=[0, 0.5])`).


##########################################################################
# References:
# -------------------------
# .. [Grotheer2021] Grotheer, Mareike, Mona Rosenke, Hua Wu, Holly Kular,
#                   Francesca R. Querdasi, Vaidehi S. Natu, Jason D. Yeatman,
#                   and Kalanit Grill-Spector. "White matter myelination during
#                   early infancy is linked to spatial gradients and myelin
#                   content at birth." Nature communications 13: 997.
#
# .. [Grotheer2023] Grotheer, Mareike, David Bloom, John Kruper,
#                   Adam Richie-Halford, Stephanie Zika,
#                   Vicente A. Aguilera González, Jason D. Yeatman,
#                   Kalanit Grill-Spector, and Ariel Rokem. "Human white matter
#                   myelinates faster in utero than ex utero." Proceedings
#                   of the National Academy of Sciences 120: e2303491120.



import numpy.testing as npt

import AFQ.data.fetch as afd
from AFQ.utils.conversion import matlab_tractography, matlab_mori_groups

import os.path as op

DATA_PATH = op.join(op.abspath(op.dirname(__file__)), "../../tests/data")


def test_matlab_tractography():
    sft = matlab_tractography(
        op.join(DATA_PATH, "WholeBrainFG_test.mat"),
        afd.read_mni_template())
    npt.assert_equal(len(sft.streamlines), 2)


def test_matlab_mori_groups():
    fiber_groups = matlab_mori_groups(
        op.join(DATA_PATH, "MoriGroups_Test.mat"),
        afd.read_mni_template())
    npt.assert_equal(len(fiber_groups.keys()), 18)
    npt.assert_equal(len(fiber_groups['Right Corticospinal'].streamlines), 2)



import numpy as np
import os.path as op
from time import time

import logging

from dipy.io.stateful_tractogram import StatefulTractogram, Space
from dipy.io.streamline import save_tractogram
import dipy.tracking.streamlinespeed as dps
import dipy.tracking.streamline as dts
from dipy.tracking.distances import bundles_distances_mdf


from AFQ.definitions.mapping import ConformedFnirtMapping


logger = logging.getLogger('AFQ')


def flip_sls(select_sl, idx_to_flip, in_place=False):
    """
    Helper function to flip streamlines
    """
    if in_place:
        flipped_sl = select_sl
    else:
        flipped_sl = [None] * len(select_sl)
    for ii, sl in enumerate(select_sl):
        if idx_to_flip[ii]:
            flipped_sl[ii] = sl[::-1]
        else:
            flipped_sl[ii] = sl
    return flipped_sl


def cut_sls_by_dist(select_sl, roi_dists, roi_idxs,
                    in_place=False):
    """
    Helper function to cut streamlines according to which points
    are closest to certain rois.

    Parameters
    ----------
    select_sl, streamlines to cut
    roi_dists, distances from a given streamline to a given inclusion roi
    roi_idxs, two indices into the list of inclusion rois to use for the cut
    in_place, whether to modify select_sl
    """
    if in_place:
        cut_sl = select_sl
    else:
        cut_sl = [None] * len(select_sl)

    for idx, this_sl in enumerate(select_sl):
        if roi_idxs[0] == -1:
            min0 = 0
        else:
            min0 = int(roi_dists[idx, roi_idxs[0]])
        if roi_idxs[1] == -1:
            min1 = len(this_sl)
        else:
            min1 = int(roi_dists[idx, roi_idxs[1]])

        # handle if sls not flipped
        if min0 > min1:
            min0, min1 = min1, min0

        # If the point that is closest to the first ROI
        # is the same as the point closest to the second ROI,
        # include the surrounding points to make a streamline.
        if min0 == min1:
            min1 = min1 + 1
            min0 = min0 - 1

        cut_sl[idx] = this_sl[min0:min1]

    return cut_sl


def read_tg(tg, nb_streamlines=None):
    if nb_streamlines and len(tg) > nb_streamlines:
        tg = StatefulTractogram.from_sft(
            dts.select_random_set_of_streamlines(
                tg.streamlines,
                nb_streamlines
            ),
            tg)
    return tg


def orient_by_streamline(sls, template_sl):
    DM = bundles_distances_mdf(
        sls,
        [template_sl, template_sl[::-1]])
    return DM[:, 0] > DM[:, 1]


def move_streamlines(tg, to, mapping, img, save_intermediates=None):
    """Move streamlines to or from template space.

    to : str
        Either "template" or "subject".
    mapping : ConformedMapping
        Mapping to use to move streamlines.
    img : Nifti1Image
        Space to move streamlines to.
    """
    tg_og_space = tg.space
    if isinstance(mapping, ConformedFnirtMapping):
        if to != "subject":
            raise ValueError(
                "Attempted to transform streamlines to template using "
                "unsupported mapping. "
                "Use something other than Fnirt.")
        tg.to_vox()
        moved_sl = []
        for sl in tg.streamlines:
            moved_sl.append(mapping.transform_inverse_pts(sl))
    else:
        tg.to_rasmm()
        if to == "template":
            volume = mapping.forward
        else:
            volume = mapping.backward
        delta = dts.values_from_volume(
            volume,
            tg.streamlines, np.eye(4))
        moved_sl = dts.Streamlines(
            [d + s for d, s in zip(delta, tg.streamlines)])
    moved_sft = StatefulTractogram(
        moved_sl,
        img,
        Space.RASMM)
    if save_intermediates is not None:
        save_tractogram(
            moved_sft,
            op.join(save_intermediates,
                    f'sls_in_{to}.trk'),
            bbox_valid_check=False)
    tg.to_space(tg_og_space)
    return moved_sft


def resample_tg(tg, n_points):
    # reformat for dipy's set_number_of_points
    if isinstance(tg, np.ndarray):
        if len(tg.shape) > 2:
            streamlines = tg.tolist()
            streamlines = [np.asarray(item) for item in streamlines]
    elif hasattr(tg, "streamlines"):
        streamlines = tg.streamlines
    else:
        streamlines = tg

    return dps.set_number_of_points(streamlines, n_points)


class SlsBeingRecognized:
    def __init__(self, sls, logger, save_intermediates, b_name, ref,
                 n_roi_dists):
        self.oriented_yet = False
        self.selected_fiber_idxs = np.arange(len(sls), dtype=np.uint32)
        self.sls_flipped = np.zeros(len(sls), dtype=np.bool8)
        self.logger = logger
        self.start_time = -1
        self.save_intermediates = save_intermediates
        self.b_name = b_name
        self.ref_sls = sls
        self.ref = ref
        self.n_roi_dists = n_roi_dists

    def initiate_selection(self, clean_name):
        self.start_time = time()
        self.logger.info(f"Filtering by {clean_name}")
        return np.zeros(len(self.selected_fiber_idxs), dtype=np.bool8)

    def select(self, idx, clean_name, cut=False):
        self.selected_fiber_idxs = self.selected_fiber_idxs[idx]
        self.sls_flipped = self.sls_flipped[idx]
        if hasattr(self, "roi_dists"):
            self.roi_dists = self.roi_dists[idx]
        time_taken = time() - self.start_time
        self.logger.info(
            f"After filtering by {clean_name} (time: {time_taken}s), "
            f"{len(self)} streamlines remain.")
        if self.save_intermediates is not None:
            save_tractogram(
                StatefulTractogram(
                    self.get_selected_sls(cut=cut),
                    self.ref, Space.VOX),
                op.join(self.save_intermediates,
                        f'sls_after_{clean_name}_for_{self.b_name}.trk'),
                bbox_valid_check=False)

    def get_selected_sls(self, cut=False, flip=False):
        selected_sls = self.ref_sls[self.selected_fiber_idxs]
        if cut and hasattr(self, "roi_dists") and self.n_roi_dists > 1:
            selected_sls = cut_sls_by_dist(
                selected_sls, self.roi_dists,
                (0, self.n_roi_dists - 1),
                in_place=False)
        if flip:
            selected_sls = flip_sls(
                selected_sls, self.sls_flipped,
                in_place=False)
        return selected_sls

    def reorient(self, idx):
        if self.oriented_yet:
            raise RuntimeError((
                "Attempted to oriented streamlines "
                "that were already oriented. "
                "This is a bug in the implementation of a "
                "bundle recognition procedure. "))
        self.oriented_yet = True
        self.sls_flipped[idx] = True

    def __bool__(self):
        return len(self) > 0

    def __len__(self):
        return len(self.selected_fiber_idxs)



import toml
import datetime
import platform
import os.path as op

from argparse import ArgumentParser
from funcargparse import FuncArgParser

from AFQ.definitions.image import *  # interprets masks loaded from toml
from AFQ.definitions.mapping import *  # interprets mappings loaded from toml
from AFQ.api.bundle_dict import *  # interprets bundle_dicts loaded from toml
from AFQ.definitions.utils import Definition
from AFQ.api.utils import kwargs_descriptors

import nibabel as nib  # allows users to input nibabel objects


def model_input_parser(usage):
    parser = ArgumentParser(usage)

    parser.add_argument("-d", "--dwi", dest='dwi', action="append",
                        help="DWI files (enter one or more)")

    parser.add_argument("-l", "--bval", dest="bval", action="append",
                        help="B-value files (enter one or more)")

    parser.add_argument("-c", "--bvec", dest="bvec", action="append",
                        help="B-vector files (enter one or more)")

    parser.add_argument("-o", "--out_dir", dest="out_dir", action="store",
                        help="""Full path to directory for files to be saved
                            (will be created if it doesn't exist)")""")

    parser.add_argument("-m", "--mask", dest="mask", action="store",
                        default=None, help="Mask file")

    parser.add_argument('-b', '--b0_threshold', dest="b0_threshold",
                        action="store", help="b0 threshold", default=0)

    return parser


def model_predict_input_parser(usage):
    parser = ArgumentParser(usage)

    parser.add_argument("-p", "--params", dest="params", action="store",
                        help="A file containing model params")

    parser.add_argument("-l", "--bval", dest="bval", action="append",
                        help="B-value files (enter one or more)")

    parser.add_argument("-c", "--bvec", dest="bvec", action="append",
                        help="B-vector files (enter one or more)")

    parser.add_argument("-o", "--out_dir", dest="out_dir", action="store",
                        help="""Full path to directory for files to be saved
                            (will be created if it doesn't exist)")""")

    parser.add_argument(
        "-s", "--S0_file", dest="S0_file", action="store",
        help="File containing S0 measurements to use in prediction")

    parser.add_argument('-b', '--b0_threshold', dest="b0_threshold",
                        help="b0 threshold (default: 0)",
                        action="store", default=0)
    return parser


def toml_to_val(t):
    if isinstance(t, str) and len(t) < 1:
        return None
    elif isinstance(t, list):
        ls = []
        for e in t:
            ls.append(toml_to_val(e))
        return ls
    elif isinstance(t, str) and t[0] == '[':
        return eval(t)
    elif isinstance(t, str) and t[0] == '{':
        return eval(t)  # interpret as dictionary
    elif isinstance(t, str) and (
            "Image" in t or "Map" in t or "Dict" in t or "_bd(" in t):
        try:
            definition_or_dict = eval(t)
        except NameError:
            return t
        if isinstance(definition_or_dict, Definition):
            return definition_or_dict
        elif isinstance(definition_or_dict, BundleDict):
            return definition_or_dict
        else:
            return t
    else:
        return t


def val_to_toml(v):
    if v is None:
        return '""'
    elif isinstance(v, Definition):
        return f'"{v.str_for_toml()}"'
    elif isinstance(v, str):
        return f'"{v}"'
    elif isinstance(v, bool):
        if v:
            return "true"
        else:
            return "false"
    elif callable(v):
        return f'"{v.__name__}"'
    elif isinstance(v, dict):
        return f'"{v}"'
    elif isinstance(v, list):
        return f'"{v}"'
    else:
        return f"{v}"


def dict_to_toml(dictionary):
    toml = '# Use \'\' to indicate None\n# Wrap dictionaries in quotes\n'
    toml = toml + '# Wrap definition object instantiations in quotes\n\n'
    for section, args in dictionary.items():
        if section == "AFQ_desc":
            toml = "# " + dictionary["AFQ_desc"].replace("\n", "\n# ")\
                + "\n\n" + toml
            continue
        toml = toml + f'[{section}]\n'
        for arg, arg_info in args.items():
            toml = toml + '\n'
            if isinstance(arg_info, dict):
                if 'desc' in arg_info:
                    toml = toml + arg_info['desc']
                toml = toml + f"{arg} = {val_to_toml(arg_info['default'])}\n"
            else:
                toml = toml + f"{arg} = {val_to_toml(arg_info)}\n"
        toml = toml + '\n'
    return toml + '\n'


# these params are handled internally in the qsiprep pipeline,
# not shown to the user (mostly BIDS filters stuff)
qsi_prep_ignore_params = [
    "bids_path", "bids_filters", "preproc_pipeline",
    "participant_labels", "output_dir"]


def dict_to_json(dictionary):
    json = "                "
    local_ignore = qsi_prep_ignore_params.copy()
    for section, args in dictionary.items():
        if section == "AFQ_desc":
            continue
        for arg, arg_info in args.items():
            if arg in local_ignore:
                continue
            local_ignore.append(arg)
            if isinstance(arg_info, dict):
                json = json\
                    + f'"{arg}": {val_to_toml(arg_info["default"])}'
            else:
                json = json + f'"{arg}": {val_to_toml(arg_info)}'
            json = json + ',\n                '
    return json[:-18]  # remove trailing ,\n and indent


def func_dict_to_arg_dict(func_dict=None, logger=None):
    if func_dict is None:
        from AFQ.recognition.recognize import recognize
        from AFQ.recognition.cleaning import clean_bundle
        import AFQ.tractography.tractography as aft
        from AFQ.api.group import GroupAFQ

        func_dict = {
            "BIDS": GroupAFQ.__init__,
            "Tractography": aft.track,
            "Segmentation": recognize,
            "Cleaning": clean_bundle}

    arg_dict = {}
    for name, func in func_dict.items():
        docstr_parser = FuncArgParser()
        docstr_parser.setup_args(func)
        if name == "BIDS":
            arg_dict["AFQ_desc"] = docstr_parser.description
        for arg, info in docstr_parser.unfinished_arguments.items():
            try:
                section = name.upper() + "_PARAMS"
                desc = info['help']
                if name != "BIDS" and 'positional' in info and info[
                        'positional']:
                    continue
            except (KeyError, IndexError) as error:
                if logger is not None:
                    logger.error(
                        "We are missing a valid description for the "
                        + f"{name} argument {arg}")
                raise error
            if section not in arg_dict.keys():
                arg_dict[section] = {}
            arg_dict[section][arg] = {}
            if 'default' in info:
                default = info['default']
            else:
                default = None
            arg_dict[section][arg]['default'] = default
            arg_dict[section][arg]['desc'] = desc

    for section, arg_info in kwargs_descriptors.items():
        section = section.upper()
        if section not in arg_dict.keys():
            arg_dict[section] = {}
        for arg, info in arg_info.items():
            if arg not in [
                    "segmentation_params",
                    "tracking_params"]:
                arg_dict[section][arg] = info

    for section, arg_info in arg_dict.items():
        if section == "AFQ_desc":
            continue
        for arg, info in arg_info.items():
            desc = arg_dict[section][arg]['desc']
            arg_dict[section][arg]['desc'] = ''
            for desc_line in desc.splitlines():
                f_desc_line = '# ' + desc_line.strip() + '\n'
                arg_dict[section][arg]['desc'] = \
                    arg_dict[section][arg]['desc'] + f_desc_line

    return arg_dict


def parse_config_run_afq(toml_file, default_arg_dict, to_call="export_all",
                         overwrite=False,
                         logger=None,
                         verbose=False,
                         dry_run=False,
                         special_args={
                             "SEGMENTATION_PARAMS": "segmentation_params",
                             "TRACTOGRAPHY_PARAMS": "tracking_params"}):
    from AFQ.api.group import GroupAFQ
    from AFQ import __version__
    # load configuration file
    if not op.exists(toml_file):
        raise FileExistsError(
            "Config file does not exist. "
            + "If you want to generate this file,"
            + " add the argument --generate-config-only")
    f_arg_dict = toml.load(toml_file)

    # extract arguments from file
    kwargs = {}
    bids_path = ''
    for section, args in f_arg_dict.items():
        for arg, default in args.items():
            if section not in default_arg_dict:
                default_arg_dict[section] = {}
            if arg == 'bids_path':
                bids_path = default
            else:
                val = toml_to_val(default)
                is_special = False
                for toml_key, doc_arg in special_args.items():
                    if section == toml_key:
                        if doc_arg not in kwargs:
                            kwargs[doc_arg] = {}
                        kwargs[doc_arg][arg] = val
                        is_special = True
                if not is_special:
                    kwargs[arg] = val
            if arg not in default_arg_dict[section]:
                default_arg_dict[section][arg] = {}
            default_arg_dict[section][arg]['default'] = default

    if logger is not None and (verbose or dry_run):
        logger.info("The following arguments are recognized: " + str(kwargs))

    if dry_run:
        return

    # if overwrite, write new file with updated docs / args
    if overwrite:
        if logger is not None:
            logger.info("Updating configuration file.")
        with open(toml_file, 'w') as ff:
            ff.write(dict_to_toml(default_arg_dict))

    if bids_path == '':
        raise RuntimeError("Config file must provide bids_path")

    # generate metadata file for this run
    default_arg_dict['pyAFQ'] = {}
    default_arg_dict['pyAFQ']['utc_time_started'] = \
        datetime.datetime.now().isoformat('T')
    default_arg_dict['pyAFQ']['version'] = __version__
    default_arg_dict['pyAFQ']['platform'] = platform.system()

    myafq = GroupAFQ(bids_path, **kwargs)

    afq_metadata_file = op.join(myafq.afq_path, 'afq_metadata.toml')
    with open(afq_metadata_file, 'w') as ff:
        ff.write(dict_to_toml(default_arg_dict))

    # call user specified function:
    if to_call == "all":
        myafq.export_all()
    else:
        myafq.export(to_call)

    # If you got this far, you can report on time ended and record that:
    default_arg_dict['pyAFQ']['utc_time_ended'] = datetime.datetime.now(
    ).isoformat('T')
    with open(afq_metadata_file, 'w') as ff:
        ff.write(dict_to_toml(default_arg_dict))


def generate_config(toml_file, default_arg_dict, overwrite=False,
                    logger=None):
    if not overwrite and op.exists(toml_file):
        raise FileExistsError(
            "Config file already exists. "
            + "If you want to overwrite this file,"
            + " add the argument --overwrite-config")
    if logger is not None:
        logger.info("Generating default configuration file.")
    toml_file = open(toml_file, 'w')
    toml_file.write(dict_to_toml(default_arg_dict))
    toml_file.close()


def generate_json(json_folder, overwrite=False,
                  logger=None):
    json_file_our_trk = op.join(json_folder, "pyafq.json")
    json_file_their_trk = op.join(json_folder, "pyafq_input_trk.json")
    if not overwrite and (
            op.exists(json_file_our_trk) or op.exists(json_file_their_trk)):
        raise FileExistsError(
            "Config file already exists. "
            + "If you want to overwrite this file,"
            + " add the argument --overwrite-config")
    if logger is not None:
        logger.info("Generating pyAFQ full pipeline QSIprep json file.")
    qsi_spec_intro_our_trk = """{
    "description": "Use pyAFQ to perform the full Tractometry pipeline",
    "space": "T1w",
    "name": "pyAFQ_full",
    "atlases": [],
    "nodes": [
        {
            "name": "pyAFQ_full",
            "software": "pyAFQ",
            "action": "pyAFQ_full",
            "input": "qsiprep",
            "output_suffix": "PYAFQ_FULL",
            "parameters": {
                "use_external_tracking": false,
                "export": "all",
"""
    qsi_spec_intro_their_trk = """{
    "description": "Use pyAFQ to perform the Tractometry pipeline, with tractography from qsiprep",
    "space": "T1w",
    "name": "pyAFQ_import_trk",
    "atlases": [],
    "nodes": [
        {
            "name": "msmt_csd",
            "software": "MRTrix3",
            "action": "csd",
            "output_suffix": "msmtcsd",
            "input": "qsiprep",
            "parameters": {
                "mtnormalize": true,
                "response": {
                "algorithm": "dhollander"
                },
                "fod": {
                "algorithm": "msmt_csd",
                "max_sh": [4, 8, 8]
                }
            }
        },
        {
            "name": "track_ifod2",
            "software": "MRTrix3",
            "action": "tractography",
            "output_suffix": "ifod2",
            "input": "msmt_csd",
            "parameters": {
                "use_5tt": false,
                "use_sift2": true,
                "tckgen":{
                "algorithm": "iFOD2",
                "select": 1e6,
                "maxlen": 250,
                "minlen": 30,
                "power":0.33
                },
                "sift2":{}
            }
        },
        {
            "name": "pyAFQ_full",
            "software": "pyAFQ",
            "action": "pyAFQ_full",
            "input": "track_ifod2",
            "output_suffix": "PYAFQ_FULL_ET",
            "parameters": {
                "use_external_tracking": true,
                "export": "all",
"""  # noqa
    qsi_spec_outro = """
            }
        }
    ]
}"""
    from AFQ.recognition.recognize import recognize
    from AFQ.recognition.cleaning import clean_bundle
    import AFQ.tractography.tractography as aft

    func_dict = {
        "Tractography": aft.track,
        "Segmentation": recognize,
        "Cleaning": clean_bundle}

    arg_dict = func_dict_to_arg_dict(func_dict, logger=logger)

    json_file = open(json_file_our_trk, 'w')
    json_file.write(qsi_spec_intro_our_trk)
    json_file.write(dict_to_json(arg_dict))
    json_file.write(qsi_spec_outro)
    json_file.close()

    json_file = open(json_file_their_trk, 'w')
    json_file.write(qsi_spec_intro_their_trk)
    json_file.write(dict_to_json(arg_dict))
    json_file.write(qsi_spec_outro)
    json_file.close()



from scipy.linalg import blas
import numpy as np

from dipy.data import default_sphere
from dipy.reconst.shm import anisotropic_power, sh_to_sf_matrix

__all__ = ["extract_odf", "anisotropic_index", "anisotropic_power"]


def extract_odf(odf, sphere=default_sphere, sh_order=8):
    """
    Calculates spherical harmonics coefficients and
    isotropic and anisotropic diffusion components
    from an ODF. Could include GFA in future updates.
    """
    odf_norm = odf / odf.max()
    ASO = odf_norm.max(axis=-1)
    ISO = odf_norm.min(axis=-1)

    _, invB = sh_to_sf_matrix(
        sphere, sh_order=sh_order, basis_type=None, return_inv=True)
    shm = blas.dgemm(
        alpha=1.,
        a=odf.reshape(-1, invB.shape[0]), b=invB).reshape(
        (*odf.shape[:-1], invB.shape[1]))

    return shm, ASO, ISO


def anisotropic_index(shm):
    """
    Calculates anisotropic index based on spherical harmonics coefficients.

    Code from Dmipy
    ---------------
    The MIT License (MIT)

    Copyright (c) 2017 Rutger Fick & Demian Wassermann

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

    https://github.com/AthenaEPI/dmipy

    References
    ----------
    .. [1] Jespersen, Sune N., et al. "Modeling dendrite density from
        magnetic resonance diffusion measurements." Neuroimage 34.4 (2007):
        1473-1486.
    """
    sh_0 = shm[..., 0] ** 2
    sh_sum_squared = np.sum(shm ** 2, axis=-1)
    AI = np.zeros_like(sh_0)
    AI = np.sqrt(1 - sh_0 / sh_sum_squared)
    return AI



"""
==========================
AFQ with HCP data
==========================
This example demonstrates how to use the AFQ API to analyze HCP data.
For this example to run properly, you will need to gain access to the HCP data.
This can be done by following this instructions on the webpage
`here <https://wiki.humanconnectome.org/display/PublicData/How+To+Connect+to+Connectome+Data+via+AWS>`_.
We will use the ``Cloudknot`` library to run our AFQ analysis in the AWS
Batch service (see also
`this example <http://yeatmanlab.github.io/pyAFQ/auto_examples/cloudknot_example.html>`_).
In the following we will use ``Cloudknot`` to run multiple
configurations of pyAFQ on the HCP dataset. Specifically, here we will run
pyAFQ with different tractography seeding strategies.
"""

##########################################################################
# Import cloudknot and set the correct region. The HCP data is stored in `us-east-1`, so it's best
# to analyze it there.
import configparser
import itertools
import cloudknot as ck
import os.path as op

ck.set_region('us-east-1')

##########################################################################
# Define a function to run. This function allows us to pass in the subject ID for the subjects we would
# like to analyze , as well as strategies for seeding tractography (different masks and/or different
# numbers of seeds per voxel).


def afq_process_subject(subject, seed_mask, n_seeds,
                        aws_access_key, aws_secret_key):
    # define a function that each job will run
    # In this case, each process does a single subject
    import logging
    import s3fs
    # all imports must be at the top of the function
    # cloudknot installs the appropriate packages from pip
    from AFQ.data.fetch import fetch_hcp
    from AFQ.api.group import GroupAFQ
    import AFQ.definitions.image as afm

    # set logging level to your choice
    logging.basicConfig(level=logging.INFO)
    log = logging.getLogger(__name__)

    # Download the given subject to the AWS Batch machine from s3
    _, hcp_bids = fetch_hcp(
        [subject],
        profile_name=False,
        study="HCP_1200",
        aws_access_key_id=aws_access_key,
        aws_secret_access_key=aws_secret_key)

    # We make a new seed mask for each process based off of the
    # seed_mask argument, which is a string.
    # This is to avoid any complications with pickling the masks.
    if seed_mask == "roi":
        seed_mask_obj = afm.RoiImage()
    elif seed_mask == "fa":
        seed_mask_obj = afm.ScalarImage("dti_fa")
    else:
        seed_mask_obj = afm.FullImage()

    # Determined if n_seeds is per voxel or not
    if n_seeds > 3:
        random_seeds = True
    else:
        random_seeds = False

    # set the tracking_params based off our inputs
    tracking_params = {
        "seed_mask": seed_mask_obj,
        "n_seeds": n_seeds,
        "random_seeds": random_seeds}

    # use segmentation file from HCP to get a brain mask,
    # where everything not labelled 0 is considered a part of the brain
    brain_mask_definition = afm.LabelledImageFile(
        suffix='seg', filters={'scope': 'dmriprep'},
        exclusive_labels=[0])

    # define the api GroupAFQ object
    myafq = GroupAFQ(
        hcp_bids,
        brain_mask_definition=brain_mask_definition,
        tracking_params=tracking_params)

    # export_all runs the entire pipeline and creates many useful derivates
    myafq.export_all()

    # upload the results to some location on s3
    myafq.upload_to_s3(
        s3fs.S3FileSystem(),
        (f"my_study_bucket/my_study_prefix_{seed_mask}_{n_seeds}"
         f"/derivatives/afq"))


##########################################################################
# In this example, we will process the data from the following subjects
subjects = ["103818", "105923", "111312"]

##########################################################################
# We will test combinations of different conditions:
# subjects, seed masks, and number of seeds
seed_mask = ["fa", "roi"]
n_seeds = [1, 2, 1000000, 2000000]

##########################################################################
# The following function creates all the combinations of the above lists, such that every subject is
# run with every mask and every number of seeds.
args = list(itertools.product(subjects, seed_mask, n_seeds))

##########################################################################
# We assume that the credentials for HCP usage are stored in the home directory in a
# `~/.aws/credentials` file. This is where these credentials are stored if the AWS CLI is used to
# configure the profile. We use the standard lib ``configparser`` library
# to get the relevant hcp keys from there.
CP = configparser.ConfigParser()
CP.read_file(open(op.join(op.expanduser('~'), '.aws', 'credentials')))
CP.sections()
aws_access_key = CP.get('hcp', 'AWS_ACCESS_KEY_ID')
aws_secret_key = CP.get('hcp', 'AWS_SECRET_ACCESS_KEY')

##########################################################################
# The following function will attach your AWS keys to each list in a list of lists
# We use this with each list being a list of arguments,
# and we append the AWS keys to each list of arguments, so that we can pass
# them into the function to be used on AWS Batch to download the data into the
# AWS Batch machines.


def attach_keys(list_of_arg_lists):
    new_list_of_arg_lists = []
    for args in list_of_arg_lists:
        arg_ls = list(args)
        arg_ls.extend([aws_access_key, aws_secret_key])
        new_list_of_arg_lists.append(arg_ls)
    return new_list_of_arg_lists


##########################################################################
# This calls the function to attach the access keys to the argument list
args = attach_keys(args)

##########################################################################
# Define the :meth:`Knot` object to run your jobs on. See
# `this example <http://yeatmanlab.github.io/pyAFQ/auto_examples/cloudknot_example.html>`_ for more
# details about the arguments to the object.
knot = ck.Knot(
    name='afq-hcp-tractography-201110-0',
    func=afq_process_subject,
    base_image='python:3.8',
    image_github_installs="https://github.com/yeatmanlab/pyAFQ.git",
    pars_policies=('AmazonS3FullAccess',),
    bid_percentage=100)

##########################################################################
# This launches a process for each combination.
# Because `starmap` is `True`, each list in `args` will be unfolded
# and passed into `afq_process_subject` as arguments.
result_futures = knot.map(args, starmap=True)

##########################################################################
# The following function can be called repeatedly in a jupyter notebook
# to view the progress of jobs::
#
#     knot.view_jobs()
#
# You can also view the status of a specific job::
#
#     knot.jobs[0].status

##########################################################################
# When all jobs are finished, remember to clobber the knot to destroy all the resources that were
# created in AWS.
result_futures.result()  # waits for futures to resolve, not needed in notebook
knot.clobber(clobber_pars=True, clobber_repo=True, clobber_image=True)

##########################################################################
# We continue processing to create another knot which takes the resulting profiles of each
# combination and combines them all into one csv file


def afq_combine_profiles(seed_mask, n_seeds):
    from AFQ.api import download_and_combine_afq_profiles
    download_and_combine_afq_profiles(
        "my_study_bucket", f"my_study_prefix_{seed_mask}_{n_seeds}")


knot2 = ck.Knot(
    name='afq_combine_subjects-201110-0',
    func=afq_combine_profiles,
    base_image='python:3.8',
    image_github_installs="https://github.com/yeatmanlab/pyAFQ.git",
    pars_policies=('AmazonS3FullAccess',),
    bid_percentage=100)

##########################################################################
# the arguments to this call to :meth:`map` are all the different configurations of pyAFQ that we ran
seed_mask = ["fa", "roi"]
n_seeds = [1, 2, 1000000, 2000000]
args = list(itertools.product(seed_mask, n_seeds))

result_futures2 = knot2.map(args, starmap=True)
result_futures2.result()
knot2.clobber(clobber_pars=True, clobber_repo=True, clobber_image=True)



def _no_tqdm(iterable, *args, **kwargs):
    """
    replacement for tqdm that just passes back the iterable to silence `tqdm`
    """
    return iterable


def _no_progressbar(progress, total_length):
    """
    no-op calls to update the `fetcher` progress bar
    """
    return


def reset_progressbars(gallery_conf, fname):
    """
    monkey patch to disable various progress bar output for examples. the
    progress bar updates pollutes sphinx gallery output. using this monkey
    patch from the spinx-build will leave the progress bars in place for other
    uses.
    """

    # disable tqdm
    import AFQ._fixes as fixes
    import AFQ.viz.utils as utils

    fixes.tqdm = _no_tqdm
    utils.tqdm = _no_tqdm

    # disable update_progressbar
    from dipy.data import fetcher

    fetcher.update_progressbar = _no_progressbar



import numpy as np
import logging

from scipy.stats import zscore

import dipy.tracking.streamline as dts
from dipy.io.stateful_tractogram import StatefulTractogram, Space

import AFQ.recognition.utils as abu
from AFQ._fixes import gaussian_weights


logger = logging.getLogger('AFQ')


def clean_by_orientation(streamlines, primary_axis, tol=None):
    """
    Compute the cardinal orientation of each streamline

    Parameters
    ----------
    streamlines : sequence of N by 3 arrays
        Where N is number of nodes in the array, the collection of
        streamlines to filter down to.

    Returns
    -------
    cleaned_idx, indicies of streamlines that passed cleaning
    """
    axis_diff = np.zeros((len(streamlines), 3))
    endpoint_diff = np.zeros((len(streamlines), 3))
    for ii, sl in enumerate(streamlines):
        # endpoint diff is between first and last
        endpoint_diff[ii, :] = np.abs(sl[0, :] - sl[-1, :])
        # axis diff is difference between the nodes, along
        axis_diff[ii, :] = np.sum(np.abs(np.diff(sl, axis=0)), axis=0)

    orientation_along = np.argmax(axis_diff, axis=1)
    along_accepted_idx = orientation_along == primary_axis
    if tol is not None:
        percentage_primary = 100 * axis_diff[:, primary_axis] / np.sum(
            axis_diff, axis=1)
        logger.debug((
            "Maximum primary percentage found: "
            f"{np.max(percentage_primary)}"))
        along_accepted_idx = np.logical_and(
            along_accepted_idx, percentage_primary > tol)

    orientation_end = np.argmax(endpoint_diff, axis=1)
    end_accepted_idx = orientation_end == primary_axis

    cleaned_idx = np.logical_and(
        along_accepted_idx,
        end_accepted_idx)

    return cleaned_idx


def clean_bundle(tg, n_points=100, clean_rounds=5, distance_threshold=3,
                 length_threshold=4, min_sl=20, stat='mean',
                 return_idx=False):
    """
    Clean a segmented fiber group based on the Mahalnobis distance of
    each streamline

    Parameters
    ----------
    tg : StatefulTractogram class instance or ArraySequence
        A whole-brain tractogram to be segmented.
    n_points : int, optional
        Number of points to resample streamlines to.
        Default: 100
    clean_rounds : int, optional.
        Number of rounds of cleaning based on the Mahalanobis distance from
        the mean of extracted bundles. Default: 5
    distance_threshold : float, optional.
        Threshold of cleaning based on the Mahalanobis distance (the units are
        standard deviations). Default: 3.
    length_threshold: float, optional
        Threshold for cleaning based on length (in standard deviations). Length
        of any streamline should not be *more* than this number of stdevs from
        the mean length.
    min_sl : int, optional.
        Number of streamlines in a bundle under which we will
        not bother with cleaning outliers. Default: 20.
    stat : callable or str, optional.
        The statistic of each node relative to which the Mahalanobis is
        calculated. Default: `np.mean` (but can also use median, etc.)
    return_idx : bool
        Whether to return indices in the original streamlines.
        Default: False.
    Returns
    -------
    A StatefulTractogram class instance containing only the streamlines
    that have a Mahalanobis distance smaller than `clean_threshold` from
    the mean of each one of the nodes.
    """
    # Convert string to callable, if that's what you got.
    if isinstance(stat, str):
        stat = getattr(np, stat)

    if hasattr(tg, "streamlines"):
        streamlines = tg.streamlines
    else:
        streamlines = dts.Streamlines(tg)

    # We don't even bother if there aren't enough streamlines:
    if len(streamlines) < min_sl:
        logger.warning((
            "Mahalanobis cleaning halted early"
            " due to low streamline count"))
        if return_idx:
            return tg, np.arange(len(streamlines))
        else:
            return tg

    # Resample once up-front:
    fgarray = np.asarray(abu.resample_tg(streamlines, n_points))

    # Keep this around, so you can use it for indexing at the very end:
    idx = np.arange(len(fgarray))
    # get lengths of each streamline
    lengths = np.array([sl.shape[0] for sl in streamlines])
    # We'll only do this for clean_rounds
    rounds_elapsed = 0
    idx_belong = idx
    while (rounds_elapsed < clean_rounds) and (np.sum(idx_belong) > min_sl):
        # Update by selection:
        idx = idx[idx_belong]
        fgarray = fgarray[idx_belong]
        lengths = lengths[idx_belong]
        rounds_elapsed += 1

        # This calculates the Mahalanobis for each streamline/node:
        m_dist = gaussian_weights(
            fgarray, return_mahalnobis=True,
            n_points=None, stat=stat)
        logger.debug(f"Shape of fgarray: {np.asarray(fgarray).shape}")
        logger.debug(f"Shape of m_dist: {m_dist.shape}")
        logger.debug(f"Maximum m_dist: {np.max(m_dist)}")
        logger.debug((
            f"Maximum m_dist for each fiber: "
            f"{np.max(m_dist, axis=1)}"))

        length_z = zscore(lengths)
        logger.debug(f"Shape of length_z: {length_z.shape}")
        logger.debug(f"Maximum length_z: {np.max(length_z)}")
        logger.debug((
            "length_z for each fiber: "
            f"{length_z}"))

        if not (
                np.any(m_dist > distance_threshold)
                or np.any(length_z > length_threshold)):
            break
        # Select the fibers that have Mahalanobis smaller than the
        # threshold for all their nodes:
        idx_dist = np.all(m_dist < distance_threshold, axis=-1)
        idx_len = length_z < length_threshold
        idx_belong = np.logical_and(idx_dist, idx_len)

        if np.sum(idx_belong) < min_sl:
            # need to sort and return exactly min_sl:
            idx_belong = np.argsort(np.sum(
                m_dist, axis=-1))[:min_sl].astype(int)
            logger.debug((
                f"At rounds elapsed {rounds_elapsed}, "
                "minimum streamlines reached"))
        else:
            idx_removed = idx_belong == 0
            logger.debug((
                f"Rounds elapsed: {rounds_elapsed}, "
                f"num removed: {np.sum(idx_removed)}"))
            logger.debug(f"Removed indicies: {np.where(idx_removed)[0]}")

    # Select based on the variable that was keeping track of things for us:
    if hasattr(tg, "streamlines"):
        out = StatefulTractogram(tg.streamlines[idx], tg, tg.space)
    else:
        out = streamlines[idx]
    if return_idx:
        return out, idx
    else:
        return out



"""
===========================================================
How to add new bundles into pyAFQ(Optic Radiations Example)
===========================================================

pyAFQ is designed to be customizable and extensible. This example shows how you
can customize it to define a new bundle based on a definition of waypoint and
endpoint ROIs of your design.

In this case, we add the optic radiations, based on work by Caffara et al. [1]_,
[2]_. The optic radiations (OR) are the primary projection of visual information
from the lateral geniculate nucleus of the thalamus to the primary visual
cortex. Studying the optic radiations with dMRI provides a linkage between white
matter tissue properties, visual perception and behavior, and physiological
responses of the visual cortex to visual stimulation.

We start by importing some of the components that we need for this example and
fixing the random seed for reproducibility

"""

import os.path as op
import plotly
import numpy as np
import shutil

from AFQ.api.group import GroupAFQ
import AFQ.api.bundle_dict as abd
import AFQ.data.fetch as afd
from AFQ.definitions.image import ImageFile, RoiImage
import AFQ.utils.streamlines as aus
np.random.seed(1234)


#############################################################################
# Get dMRI data
# ---------------
# We will analyze one subject from the Healthy Brain Network Processed Open
# Diffusion Derivatives dataset (HBN-POD2) [3]_, [4]_. We'll use a fetcher to
# get preprocessed dMRI data for one of the >2,000 subjects in that study. The
# data gets organized into a BIDS-compatible format in the `~/AFQ_data/HBN`
# folder:

study_dir = afd.fetch_hbn_preproc(["NDARAA948VFH"])[1]

#############################################################################
# Define custom `BundleDict` object
# --------------------------------
# The `BundleDict` object holds information about "include" and "exclude" ROIs,
# as well as endpoint ROIS, and whether the bundle crosses the midline. In this
# case, the ROIs are all defined in the MNI template space that is used as the
# default template space in pyAFQ, but, in principle, other template spaces
# could be used.
#
# The ROIs for the case can be downloaded using a custom fetcher which saves
# the ROIs to a folder and creates a dictionary of paths to the ROIs:

or_rois = afd.read_or_templates()

bundles = abd.OR_bd()

#############################################################################
# Custom bundle definitions such as the OR, and the standard BundleDict can be
# combined through addition. To get both the OR and the standard bundles, we
# would execute the following code::
#
#     bundles = bundles + abd.default18_bd()
#
# In this case, we will skip this and generate just the OR.


#############################################################################
# Define GroupAFQ object
# ----------------------
# HBN POD2 have been processed with qsiprep [5]_. This means that a brain mask
# has already been computer for them. As you can see in other examples, these
# data also have a mapping calculated for them, which can also be incorporated
# into processing. However, in this case, we will let pyAFQ calculate its own
# SyN-based mapping so that the `combine_bundle` method can be used below to
# create a montage visualization.
#
# For tractography, we use CSD-based probabilistic tractography seeding
# extensively (`n_seeds=4` means 81 seeds per voxel!), but only within the ROIs
# and not throughout the white matter. This is controlled by passing
# `"seed_mask": RoiImage()` in the `tracking_params` dict. The custom bundles
# are passed as `bundle_info=bundles`. The call to `my_afq.export_all()`
# initiates the pipeline.

brain_mask_definition = ImageFile(
    suffix="mask",
    filters={'desc': 'brain',
             'space': 'T1w',
             'scope': 'qsiprep'})

my_afq = GroupAFQ(
    bids_path=study_dir,
    preproc_pipeline="qsiprep",
    output_dir=op.join(study_dir, "derivatives", "afq_or"),
    brain_mask_definition=brain_mask_definition,
    tracking_params={"n_seeds": 4,
                     "directions": "prob",
                     "odf_model": "CSD",
                     "seed_mask": RoiImage()},
    bundle_info=bundles)

my_afq.export_all()

#############################################################################
# Visualize a montage
# ----------------------
# One way to examine the output of the pyAFQ pipeline is by creating a montage
# of images of a particular bundle across a group of participants (or, in this
# case, the one participant that was analyzed).
#
# .. note::
#
#   The montage file is copied to the present working directory so that it gets
#   properly rendered into the web-page containing this example. It is not
#   necessary to do this when running this type of analysis.

my_afq.combine_bundle("Left Optic Radiation")
montage = my_afq.group_montage(
    "Left Optic Radiation",
    (1, 1), "Axial", "left")
shutil.copy(montage[0], op.split(montage[0])[-1])

#############################################################################
# Interactive bundle visualization
# --------------------------------
# Another way to examine the outputs is to export the individual bundle
# figures, which show the streamlines, as well as the ROIs used to define the
# bundle. This is an html file, which contains an interactive figure that can
# be navigated, zoomed, rotated, etc.

bundle_html = my_afq.export("indiv_bundles_figures")
plotly.io.show(bundle_html["NDARAA948VFH"]["Left Optic Radiation"])

#############################################################################
# References
# ----------
# .. [1] Caffarra S, Joo SJ, Bloom D, Kruper J, Rokem A, Yeatman JD. Development
#     of the visual white matter pathways mediates development of
#     electrophysiological responses in visual cortex. Hum Brain Mapp.
#     2021;42(17):5785-5797.
#
# .. [2] Caffarra S, Kanopka K, Kruper J, Richie-Halford A, Roy E, Rokem A,
#     Yeatman JD. Development of the alpha rhythm is linked to visual white
#     matter pathways and visual detection performance. bioRxiv.
#     doi:10.1101/2022.09.03.506461
#
# .. [3] Alexander LM, Escalera J, Ai L, et al. An open resource for
#     transdiagnostic research in pediatric mental health and learning
#     disorders. Sci Data. 2017;4:170181.
#
# .. [4] Richie-Halford A, Cieslak M, Ai L, et al. An analysis-ready and quality
#     controlled resource for pediatric brain white-matter research. Scientific
#     Data. 2022;9(1):1-27.
#
# .. [5] Cieslak M, Cook PA, He X, et al. QSIPrep: an integrative platform for
#     preprocessing and reconstructing diffusion MRI data. Nat Methods.
#     2021;18(7):775-778.



"""
==========================
How to use Free water DTI
==========================

The free-water DTI model [1, 2]_ fits a two compartment model to dMRI data
with more than one non-zero shell. One compartment is a spherical compartment
with the diffusivity of water, which accounts for free water in the tissue.
The other compartment is the standard diffusion tensor.

In this example, we will compare the results of the fwDTI model and the
standard DTI model.

"""
import os.path as op

import matplotlib.pyplot as plt
import nibabel as nib

from AFQ.api.group import GroupAFQ
import AFQ.data.fetch as afd

from AFQ.definitions.image import ImageFile, RoiImage
import AFQ.api.bundle_dict as abd

import pandas as pd

#############################################################################
# Get some data
# --------------
# In this example, we will look at one subject from the Healthy Brain Network
# Processed Open Diffusion Derivatives dataset (HBN-POD2) [3, 4]_. The data in
# this study were collected with a multi-shell sequence, meaning that most
# subjects in this study have data with more than one non-zero b-value. This
# means that we can fit the fwDTI model to their data.
#
# We'll use a fetcher to get preprocessd dMRI data for one of the >2,000
# subjects in that study. The data gets organized into a BIDS-compatible
# format in the `~/AFQ_data/HBN` folder.

study_dir = afd.fetch_hbn_preproc(["NDARAA948VFH"])[1]

#############################################################################
# Define an AFQ object
# --------------------
# In addition to preprocessd dMRI data, HBN-POD2 contains brain mask and mapping
# information for each subject. We can use this information in our pipeline, by
# inserting this information as `mapping_definition` and `brain_mask_definition`
# inputs to the `GroupAFQ` class initializer. When initializing this object, we
# will also ask for the fwDTI scalars to be computed. For expedience, we will
# limit our investigation to the bilateral arcuate fasciculus and track only
# around that bundle. If you would like to do this for all bundles, you would
# remove the `bundle_dict` and `tracking_params` inputs to the initializer that
# are provided below.

brain_mask_definition = ImageFile(
    suffix="mask",
    filters={'desc': 'brain',
             'space': 'T1w',
             'scope': 'qsiprep'})

bundle_names = ["Left Arcuate", "Right Arcuate"]
bundle_dict = abd.default18_bd()[bundle_names]

myafq = GroupAFQ(
    bids_path=study_dir,
    preproc_pipeline='qsiprep',
    output_dir=op.join(study_dir, "derivatives", "afq_fwdti"),
    bundle_info=bundle_dict,
    tracking_params={
        "n_seeds": 50000,
        "random_seeds": True,
        "seed_mask": RoiImage(use_waypoints=True, use_endpoints=True),
    },
    brain_mask_definition=brain_mask_definition,
    scalars=["fwdti_fa", "fwdti_md", "fwdti_fwf", "dti_fa", "dti_md"])

#############################################################################
# Compare fwDTI and DTI maps
# -------------------------
# First, we take a look at the maps for the FA and MD calculated using the two
# models

fwFA = nib.load(myafq.export("fwdti_fa")["NDARAA948VFH"]).get_fdata()
FA = nib.load(myafq.export("dti_fa")["NDARAA948VFH"]).get_fdata()

fig, ax = plt.subplots(1, 2)
ax[0].matshow(FA[:, :, FA.shape[-1] // 2], cmap='gray')
ax[0].axis("off")

ax[1].matshow(fwFA[:, :, fwFA.shape[-1] // 2], cmap='gray')
ax[1].axis("off")


fwMD = nib.load(myafq.export("fwdti_md")["NDARAA948VFH"]).get_fdata()
MD = nib.load(myafq.export("dti_md")["NDARAA948VFH"]).get_fdata()

fig, ax = plt.subplots(1, 2)
ax[0].matshow(MD[:, :, MD.shape[-1] // 2], cmap='gray', vmax=0.005)
ax[0].axis("off")

ax[1].matshow(fwMD[:, :, fwMD.shape[-1] // 2], cmap='gray', vmax=0.005)
ax[1].axis("off")


#############################################################################
#  Free-water fraction map
# -------------------------
# In addition to the standard tensor scalars, provided by the fwDTI model, this
# model also computes a free-water fraction, which is a number between 0 and 1
# that assesses the fraction of the voxel signal that is explained by the free
# water compartment.

fwf = nib.load(myafq.export("fwdti_fwf")["NDARAA948VFH"]).get_fdata()
fig, ax = plt.subplots()
ax.matshow(fwf[:, :, fwf.shape[-1] // 2], cmap='gray')
ax.axis("off")

#############################################################################
#  Comparing bundle profiles
# -------------------------
# Exporting the profiles will create a CSV file that contains information about
# node-by-node values of the scalars computed with both models. Here, we read in
# this information with Pandas and plot a comparison. As you can see, when free
# water is accounted for with the fwDTI model, FA along the bundle is higher and
# MD is lower than that estimated with the standard DTI model.

profiles_csv = myafq.export("profiles")['NDARAA948VFH']
profiles = pd.read_csv(profiles_csv)

fig, ax = plt.subplots(3, 2)
for ii, bundle in enumerate(["Left Arcuate", "Right Arcuate"]):
    ax[0, ii].plot(profiles[profiles["tractID"] == bundle]["fwdti_fa"],
                   label="fwDTI")
    ax[0, ii].plot(profiles[profiles["tractID"] == bundle]["dti_fa"],
                   label="DTI")
    ax[0, ii].set_ylabel("FA")
    ax[0, ii].legend()
    ax[1, ii].plot(profiles[profiles["tractID"] == bundle]["fwdti_md"],
                   label="fwDTI")
    ax[1, ii].plot(profiles[profiles["tractID"] == bundle]["dti_md"],
                   label="DTI")
    ax[1, ii].set_ylabel("MD")
    ax[1, ii].legend()
    ax[2, ii].plot(profiles[profiles["tractID"] == bundle]["fwdti_fwf"])
    ax[2, ii].set_ylabel("Free water fraction")
    ax[2, ii].set_xlabel("Distance along the bundle (A => P)")


#############################################################################
# References
# ----------
#
# .. [1] Hoy AR, Koay CG, Kecskemeti SR, Alexander AL. Optimization of a free
#     water elimination two-compartment model for diffusion tensor imaging.
#     Neuroimage. 2014;103:323-333.
#
# .. [2] Henriques RN, Rokem A, Garyfallidis E, St-Jean S, Peterson ET, Correia
#     MM. [Re] Optimization of a free water elimination two-compartment model
#     for diffusion tensor imaging. bioRxiv. February 2017:108795.
#     doi:10.1101/108795
#
# .. [3] Alexander LM, Escalera J, Ai L, et al. An open resource for
#     transdiagnostic research in pediatric mental health and learning
#     disorders. Sci Data. 2017;4:170181.
#
# .. [4] Richie-Halford A, Cieslak M, Ai L, et al. An analysis-ready and
#     quality controlled resource for pediatric brain white-matter research.
#     Scientific Data. 2022;9(1):1-27.



import numpy as np

import dipy.tracking.streamlinespeed as dps


def sl_curve(sl, n_points):
    """
    Calculate the direction of the displacement between
    each point along a streamline

    Parameters
    ----------
    sl : 2d array-like
        Streamline to calcualte displacements for.
    n_points : int
        Number of points to resample the streamline to

    Returns
    -------
    2d array of shape (len(sl)-1, 3) with displacements
    between each point in sl normalized to 1.
    """
    # Resample to a standardized number of points
    resampled_sl = dps.set_number_of_points(
        sl,
        n_points)

    # displacement at each point
    resampled_sl_diff = np.diff(resampled_sl, axis=0)

    # normalize this displacement
    resampled_sl_diff = resampled_sl_diff / np.linalg.norm(
        resampled_sl_diff, axis=1)[:, None]

    return resampled_sl_diff


def sl_curve_dist(curve1, curve2):
    """
    Calculate the mean angle using the directions of displacement
    between two streamlines

    Parameters
    ----------
    curve1, curve2 : 2d array-like
        Two curves calculated from sl_curve.

    Returns
    -------
    The mean angle between each curve across all steps, in radians
    """
    return np.mean(np.arccos(np.sum(curve1 * curve2, axis=1)))



import numpy as np
import os.path as op
import os
import logging

import dipy.tracking.streamlinespeed as dps
from dipy.io.stateful_tractogram import StatefulTractogram, Space

import AFQ.recognition.utils as abu
from AFQ.api.bundle_dict import BundleDict
from AFQ.recognition.criteria import run_bundle_rec_plan
from AFQ.recognition.preprocess import get_preproc_plan


logger = logging.getLogger('AFQ')


def recognize(
        tg,
        img,
        mapping,
        bundle_dict,
        reg_template,
        nb_points=False,
        nb_streamlines=False,
        clip_edges=False,
        parallel_segmentation={"engine": "serial"},
        rb_recognize_params=dict(
            model_clust_thr=1.25,
            reduction_thr=25,
            pruning_thr=12),
        refine_reco=False,
        prob_threshold=0,
        dist_to_waypoint=None,
        rng=None,
        return_idx=False,
        filter_by_endpoints=True,
        dist_to_atlas=4,
        save_intermediates=None,
        cleaning_params={}):
    """
    Segment streamlines into bundles.

    Parameters
    ----------
    tg : str, StatefulTractogram
        Tractogram to segment.
    img : str, nib.Nifti1Image
        Image for reference.
    mapping : MappingDefinition
        Mapping from subject to template.
    bundle_dict : dict or AFQ.api.BundleDict
        Dictionary of bundles to segment.
    reg_template : str, nib.Nifti1Image
        Template image for registration.
    nb_points : int, boolean
        Resample streamlines to nb_points number of points.
        If False, no resampling is done. Default: False
    nb_streamlines : int, boolean
        Subsample streamlines to nb_streamlines.
        If False, no subsampling is don. Default: False
    clip_edges : bool
        Whether to clip the streamlines to be only in between the ROIs.
        Default: False
    parallel_segmentation : dict or AFQ.api.BundleDict
        How to parallelize segmentation across processes when performing
        waypoint ROI segmentation. Set to {"engine": "serial"} to not
        perform parallelization. Some engines may cause errors, depending
        on the system. See ``dipy.utils.parallel.paramap`` for
        details.
        Default: {"engine": "serial"}
    rb_recognize_params : dict
        RecoBundles parameters for the recognize function.
        Default: dict(model_clust_thr=1.25, reduction_thr=25, pruning_thr=12)
    refine_reco : bool
        Whether to refine the RecoBundles segmentation.
        Default: False
    prob_threshold : float.
        Using AFQ Algorithm.
        Initial cleaning of fiber groups is done using probability maps
        from [Hua2008]_. Here, we choose an average probability that
        needs to be exceeded for an individual streamline to be retained.
        Default: 0.
    dist_to_waypoint : float.
        The distance that a streamline node has to be from the waypoint
        ROI in order to be included or excluded.
        If set to None (default), will be calculated as the
        center-to-corner distance of the voxel in the diffusion data.
        If a bundle has inc_addtol or exc_addtol in its bundle_dict, that
        tolerance will be added to this distance.
        For example, if you wanted to increase tolerance for the right
        arcuate waypoint ROIs by 3 each, you could make the following
        modification to your bundle_dict:
        bundle_dict["Right Arcuate"]["inc_addtol"] = [3, 3]
        Additional tolerances can also be negative.
        Default: None.
    rng : RandomState or int
        If None, creates RandomState.
        If int, creates RandomState with seed rng.
        Used in RecoBundles Algorithm.
        Default: None.
    return_idx : bool
        Whether to return the indices in the original streamlines as part
        of the output of segmentation.
        Default: False.
    filter_by_endpoints: bool
        Whether to filter the bundles based on their endpoints.
        Default: True.
    dist_to_atlas : float
        If filter_by_endpoints is True, this is the required distance
        from the endpoints to the atlas ROIs.
        Default: 4
    save_intermediates : str, optional
        The full path to a folder into which intermediate products
        are saved. Default: None, means no saving of intermediates.
    cleaning_params : dict, optional
        Cleaning params to pass to seg.clean_bundle. This will
        override the default parameters of that method. However, this
        can be overriden by setting the cleaning parameters in the
        bundle_dict. Default: {}.

    References
    ----------
    .. [Hua2008] Hua K, Zhang J, Wakana S, Jiang H, Li X, et al. (2008)
    Tract probability maps in stereotaxic spaces: analyses of white
    matter anatomy and tract-specific quantification. Neuroimage 39:
    336-347
    .. [Yeatman2012] Yeatman, Jason D., Robert F. Dougherty, Nathaniel J.
    Myall, Brian A. Wandell, and Heidi M. Feldman. 2012. "Tract Profiles of
    White Matter Properties: Automating Fiber-Tract Quantification"
    PloS One 7 (11): e49790.
    .. [Garyfallidis17] Garyfallidis et al. Recognition of white matter
    bundles using local and global streamline-based registration and
    clustering, Neuroimage, 2017.
    """
    if rng is None:
        rng = np.random.RandomState()
    elif isinstance(rng, int):
        rng = np.random.RandomState(rng)

    if (save_intermediates is not None) and \
            (not op.exists(save_intermediates)):
        os.makedirs(save_intermediates, exist_ok=True)

    logger.info("Preprocessing Streamlines")
    tg = abu.read_tg(tg, nb_streamlines)

    # If resampling over-write the sft:
    if nb_points:
        tg = StatefulTractogram(
            dps.set_number_of_points(tg.streamlines, nb_points),
            tg, tg.space)

    if not isinstance(bundle_dict, BundleDict):
        bundle_dict = BundleDict(bundle_dict)

    tg.to_vox()
    n_streamlines = len(tg)
    bundle_decisions = np.zeros(
        (n_streamlines, len(bundle_dict)),
        dtype=np.bool8)
    bundle_to_flip = np.zeros(
        (n_streamlines, len(bundle_dict)),
        dtype=np.bool8)
    bundle_roi_dists = -np.ones(
        (
            n_streamlines,
            len(bundle_dict),
            bundle_dict.max_includes),
        dtype=np.uint32)

    fiber_groups = {}
    meta = {}

    preproc_imap = get_preproc_plan(img, tg, dist_to_waypoint, dist_to_atlas)

    logger.info("Assigning Streamlines to Bundles")
    for bundle_idx, bundle_name in enumerate(
            bundle_dict.bundle_names):
        logger.info(f"Finding Streamlines for {bundle_name}")
        run_bundle_rec_plan(
            bundle_dict, tg, mapping, img, reg_template, preproc_imap,
            bundle_name, bundle_idx, bundle_to_flip, bundle_roi_dists,
            bundle_decisions,
            clip_edges=clip_edges,
            parallel_segmentation=parallel_segmentation,
            rb_recognize_params=rb_recognize_params,
            prob_threshold=prob_threshold,
            refine_reco=refine_reco,
            rng=rng,
            return_idx=return_idx,
            filter_by_endpoints=filter_by_endpoints,
            save_intermediates=save_intermediates,
            cleaning_params=cleaning_params)

    if save_intermediates is not None:
        os.makedirs(save_intermediates, exist_ok=True)
        bc_path = op.join(save_intermediates,
                          "sls_bundle_decisions.npy")
        np.save(bc_path, bundle_decisions)

    conflicts = np.sum(np.sum(bundle_decisions, axis=1) > 1)
    if conflicts > 0:
        logger.warning((
            "Conflicts in bundle assignment detected. "
            f"{conflicts} conflicts detected in total out of "
            f"{n_streamlines} total streamlines."
            "Defaulting to whichever bundle appears first"
            "in the bundle_dict."))
    bundle_decisions = np.concatenate((
        bundle_decisions, np.ones((n_streamlines, 1))), axis=1)
    bundle_decisions = np.argmax(bundle_decisions, -1)

    # We do another round through, so that we can:
    # 1. Clip streamlines according to ROIs
    # 2. Re-orient streamlines
    logger.info("Re-orienting streamlines to consistent directions")
    for bundle_idx, bundle in enumerate(bundle_dict.bundle_names):
        logger.info(f"Processing {bundle}")

        select_idx = np.where(bundle_decisions == bundle_idx)[0]

        if len(select_idx) == 0:
            # There's nothing here, set and move to the next bundle:
            if "bundlesection" in bundle_dict.get_b_info(bundle):
                for sb_name in bundle_dict.get_b_info(bundle)[
                        "bundlesection"]:
                    _return_empty(sb_name, return_idx, fiber_groups, img)
            else:
                _return_empty(bundle, return_idx, fiber_groups, img)
            continue

        # Use a list here, because ArraySequence doesn't support item
        # assignment:
        select_sl = list(tg.streamlines[select_idx])
        roi_dists = bundle_roi_dists[select_idx, bundle_idx, :]
        n_includes = len(bundle_dict.get_b_info(
            bundle).get("include", []))
        if clip_edges and n_includes > 1:
            logger.info("Clipping Streamlines by ROI")
            select_sl = abu.cut_sls_by_dist(
                select_sl, roi_dists,
                (0, n_includes - 1), in_place=True)

        to_flip = bundle_to_flip[select_idx, bundle_idx]
        b_def = dict(bundle_dict.get_b_info(bundle_name))
        if "bundlesection" in b_def:
            for sb_name, sb_include_cuts in bundle_dict.get_b_info(
                    bundle)["bundlesection"].items():
                bundlesection_select_sl = abu.cut_sls_by_dist(
                    select_sl, roi_dists,
                    sb_include_cuts, in_place=False)
                _add_bundle_to_fiber_group(
                    sb_name, bundlesection_select_sl, select_idx,
                    to_flip, return_idx, fiber_groups, img)
                _add_bundle_to_meta(sb_name, b_def)
        else:
            _add_bundle_to_fiber_group(
                bundle, select_sl, select_idx, to_flip,
                return_idx, fiber_groups, img)
            _add_bundle_to_meta(bundle, b_def, meta)
    return fiber_groups, meta


# Helper functions for formatting the results
def _return_empty(bundle_name, return_idx, fiber_groups, img):
    """
    Helper function to return an empty dict under
    some conditions.
    """
    if return_idx:
        fiber_groups[bundle_name] = {}
        fiber_groups[bundle_name]['sl'] = StatefulTractogram(
            [], img, Space.VOX)
        fiber_groups[bundle_name]['idx'] = np.array([])
    else:
        fiber_groups[bundle_name] = StatefulTractogram(
            [], img, Space.VOX)


def _add_bundle_to_fiber_group(b_name, sl, idx, to_flip,
                               return_idx, fiber_groups, img):
    """
    Helper function to add a bundle to a fiber group.
    """
    sl = abu.flip_sls(
        sl, to_flip,
        in_place=False)

    sl = StatefulTractogram(
        sl,
        img,
        Space.VOX)

    if return_idx:
        fiber_groups[b_name] = {}
        fiber_groups[b_name]['sl'] = sl
        fiber_groups[b_name]['idx'] = idx
    else:
        fiber_groups[b_name] = sl


def _add_bundle_to_meta(bundle_name, b_def, meta):
    # remove keys that can never be serialized
    for key in [
            'include', 'exclude', 'prob_map',
            'start', 'end', 'curvature']:
        b_def.pop(key, None)
    meta[bundle_name] = b_def



"""
=========================================
Using RecoBundles for bundle recognition
=========================================

For bundle recognition, pyAFQ defaults to use the waypoint ROI approach
described in [Yeatman2012]_. However, as an alternative approach, pyAFQ also
supports using the RecoBundles algorithm [Garyfallidis2017]_, which uses an
atlas of bundles in streamlines. This example shows how to
use RecoBundles for bundle recognition.

The code closely resembles the code used in :ref:`sphx_glr_tutorial_examples_plot_001-plot_afq_api.py`.

"""

import os.path as op
import AFQ.data.fetch as afd
from AFQ.api.group import GroupAFQ
import AFQ.api.bundle_dict as abd

afd.organize_stanford_data(clear_previous_afq="track")

tracking_params = dict(n_seeds=25000,
                       random_seeds=True,
                       rng_seed=42)


##########################################################################
# Defining the segmentation params
# --------------------------------
# We also refer to bundle recognition as the "segmentation" of the tractogram.
# Parameters of this process are set through a dictionary input to the
# `segmentation_params` argument of the GroupAFQ object. In this case, we
# use `abd.reco_bd(16)`, which tells pyAFQ to use the RecoBundles
# algorithm for bundle recognition.

myafq = GroupAFQ(
    output_dir=op.join(afd.afq_home, 'stanford_hardi', 'derivatives',
                       'recobundles'),
    bids_path=op.join(afd.afq_home, 'stanford_hardi'),
    # Set the algorithm to use RecoBundles for bundle recognition:
    bundle_info=abd.reco_bd(16),
    preproc_pipeline='vistasoft',
    tracking_params=tracking_params,
    viz_backend_spec='plotly_no_gif')

fig_files = myafq.export_all()

##########################################################################
# References:
# -------------------------
#  .. [Garyfallidis2017] Garyfallidis, Eleftherios, Marc-Alexandre Côté,
#                      Francois Rheault, Jasmeen Sidhu, Janice Hau, Laurent
#                      Petit, David Fortin, Stephen Cunanne, and Maxime
#                      Descoteaux. 2017.“Recognition of White Matter Bundles
#                      Using Local and Global Streamline-Based Registration and
#                      Clustering.”NeuroImage 170: 283-295.



import numpy as np
import logging

from scipy.special import lpmv, gammaln

from tqdm import tqdm

import math

from dipy.reconst.gqi import squared_radial_component
from dipy.data import default_sphere
from dipy.tracking.streamline import set_number_of_points
from scipy.linalg import blas


logger = logging.getLogger('AFQ')


def gwi_odf(gqmodel, data):
    gqi_vector = np.real(
        squared_radial_component(np.dot(
            gqmodel.b_vector, default_sphere.vertices.T)
            * gqmodel.Lambda))
    odf = blas.dgemm(
        alpha=1.,
        a=data.reshape(-1, gqi_vector.shape[0]),
        b=gqi_vector
    ).reshape((*data.shape[:-1], gqi_vector.shape[1]))
    return odf


def spherical_harmonics(m, n, theta, phi):
    """
    An implementation of spherical harmonics that overcomes conda compilation
    issues. See: https://github.com/nipy/dipy/issues/852
    """
    x = np.cos(phi)
    val = lpmv(m, n, x).astype(complex)
    val *= np.sqrt((2 * n + 1) / 4.0 / np.pi)
    val *= np.exp(0.5 * (gammaln(n - m + 1) - gammaln(n + m + 1)))
    val = val * np.exp(1j * m * theta)
    return val


def in_place_norm(vec, axis=-1, keepdims=False, delvec=True):
    """ Return Vectors with Euclidean (L2) norm

    See :term:`unit vector` and :term:`Euclidean norm`

    Parameters
    -------------
    vec : array_like
        Vectors to norm. Squared in the process of calculating the norm.
    axis : int, optional
        Axis over which to norm. By default norm over last axis. If `axis` is
        None, `vec` is flattened then normed. Default is -1.
    keepdims : bool, optional
        If True, the output will have the same number of dimensions as `vec`,
        with shape 1 on `axis`. Default is False.
    delvec : bool, optional
        If True, vec is deleted as soon as possible.
        If False, vec is not deleted, but still squared. Default is True.

    Returns
    ---------
    norm : array
        Euclidean norms of vectors.

    Examples
    --------
    >>> vec = [[8, 15, 0], [0, 36, 77]]
    >>> in_place_norm(vec)
    array([ 17.,  85.])
    >>> vec = [[8, 15, 0], [0, 36, 77]]
    >>> in_place_norm(vec, keepdims=True)
    array([[ 17.],
           [ 85.]])
    >>> vec = [[8, 15, 0], [0, 36, 77]]
    >>> in_place_norm(vec, axis=0)
    array([  8.,  39.,  77.])
    """
    vec = np.asarray(vec)

    if keepdims:
        ndim = vec.ndim
        shape = vec.shape

    np.square(vec, out=vec)
    vec_norm = vec.sum(axis)
    if delvec:
        del vec
    try:
        np.sqrt(vec_norm, out=vec_norm)
    except TypeError:
        vec_norm = vec_norm.astype(float)
        np.sqrt(vec_norm, out=vec_norm)

    if keepdims:
        if axis is None:
            shape = [1] * ndim
        else:
            shape = list(shape)
            shape[axis] = 1
        vec_norm = vec_norm.reshape(shape)

    return vec_norm


def tensor_odf(evals, evecs, sphere, num_batches=100):
    """
    Calculate the tensor Orientation Distribution Function

    Parameters
    ----------
    evals : array (4D)
        Eigenvalues of a tensor. Shape (x, y, z, 3).
    evecs : array (5D)
        Eigenvectors of a tensor. Shape (x, y, z, 3, 3)
    sphere : sphere object
        The ODF will be calculated in each vertex of this sphere.
    num_batches : int
        Split the calculation into batches. This reduces memory usage.
        If memory use is not an issue, set to 1.
        If set to -1, there will be 1 batch per vertex in the sphere.
        Default: 100
    """
    num_vertices = sphere.vertices.shape[0]
    if num_batches == -1:
        num_batches = num_vertices
    batch_size = math.ceil(num_vertices / num_batches)
    batches = range(num_batches)

    mask = np.where((evals[..., 0] > 0)
                    & (evals[..., 1] > 0)
                    & (evals[..., 2] > 0))
    evecs = evecs[mask]

    proj_norm = np.zeros((num_vertices, evecs.shape[0]))

    it = tqdm(batches) if num_batches != 1 else batches
    for i in it:
        start = i * batch_size
        end = (i + 1) * batch_size
        if end > num_vertices:
            end = num_vertices

        proj = np.dot(sphere.vertices[start:end], evecs)
        proj /= np.sqrt(evals[mask])
        proj_norm[start:end, :] = in_place_norm(proj)

    proj_norm **= -3
    proj_norm /= 4 * np.pi * np.sqrt(np.prod(evals[mask], -1))

    odf = np.zeros((evals.shape[:3] + (sphere.vertices.shape[0],)))
    odf[mask] = proj_norm.T
    return odf


def gaussian_weights(bundle, n_points=100, return_mahalnobis=False,
                     stat=np.mean):
    """
    Calculate weights for each streamline/node in a bundle, based on a
    Mahalanobis distance from the core the bundle, at that node (mean, per
    default).

    Parameters
    ----------
    bundle : Streamlines
        The streamlines to weight.
    n_points : int or None, optional
        The number of points to resample to. If this is None, we assume bundle
        is already resampled, and do not do any resampling. Default: 100.
    return_mahalanobis : bool, optional
        Whether to return the Mahalanobis distance instead of the weights.
        Default: False.
    stat : callable, optional.
        The statistic used to calculate the central tendency of streamlines in
        each node. Can be one of {`np.mean`, `np.median`} or other functions
        that have similar API. Default: `np.mean`
    resample : bool, optional
        Whether its necessary to resample the streamlines to the same number
        of points. Only set to False if they are already resampled.
        Default: True.
    Returns
    -------
    w : array of shape (n_streamlines, n_points)
        Weights for each node in each streamline, calculated as its relative
        inverse of the Mahalanobis distance, relative to the distribution of
        coordinates at that node position across streamlines.

    """
    if n_points is not None:
        if isinstance(bundle, np.ndarray):
            bundle = bundle.tolist()
        if isinstance(bundle, list):
            bundle = [np.asarray(item) for item in bundle]
        sls = np.asarray(set_number_of_points(bundle, n_points))
    else:
        sls = bundle

    n_sls, n_nodes, _ = sls.shape

    if n_sls < 15:  # Cov^-1 unstable under this amount
        weights = np.ones((n_sls, n_nodes))
        logger.warning((
            "Not enough streamlines for weight calculation, "
            "weighting everything evenly"))
        if return_mahalnobis:
            return np.full((n_sls, n_nodes), np.nan)
        else:
            return weights / np.sum(weights, 0)
    else:
        weights = np.zeros((n_sls, n_nodes))
    diff = stat(sls, axis=0) - sls
    for i in range(n_nodes):
        # This should come back as a 3D covariance matrix with the spatial
        # variance covariance of this node across the different streamlines,
        # reorganized as an upper diagonal matrix for expected Mahalanobis
        cov = np.cov(sls[:, i, :].T, ddof=0)
        while np.any(np.linalg.eigvals(cov) < 0):
            cov += np.eye(cov.shape[0]) * 1e-12

        # calculate Mahalanobis for node in every fiber
        if np.any(cov > 0):
            ci = np.linalg.inv(cov)

            dist = (diff[:, i, :] @ ci) * diff[:, i, :]
            weights[:, i] = np.sqrt(np.sum(dist, axis=1))

        # In the special case where all the streamlines have the exact same
        # coordinate in this node, the covariance matrix is all zeros, so
        # we can't calculate the Mahalanobis distance, we will instead give
        # each streamline an identical weight, equal to the number of
        # streamlines:
        else:
            weights[:, i] = 0
    if return_mahalnobis:
        return weights

    # weighting is inverse to the distance (the further you are, the less you
    # should be weighted)
    weights = 1 / weights
    # Normalize before returning, so that the weights in each node sum to 1:
    return weights / np.sum(weights, 0)



"""
======================================
Getting started programming with pyAFQ
======================================

There are two ways to :doc:`use pyAFQ </howto/getting_started>`: through the
command line interface, and by writing Python code. This tutorial will walk you
through the basics of the latter, using pyAFQ's Python Application Programming
Interface (API).

"""
import os.path as op

import matplotlib.pyplot as plt
import nibabel as nib
import plotly
import pandas as pd

from AFQ.api.group import GroupAFQ
import AFQ.data.fetch as afd
import AFQ.viz.altair as ava

##########################################################################
# Example data
# ------------
# pyAFQ assumes that the data is organized in a BIDS compliant directory.
# To get users started with this tutorial, we will download some example
# data and organize it in a BIDS compliant way (for more details on how
# BIDS is used in pyAFQ, refer to :ref:`bids_tutorial`).
#
# The following call dowloads a dataset that contains a single subject's
# high angular resolution diffusion imaging (HARDI) data, collected at the
# Stanford Vista Lab
#
# .. note::
#   See https://purl.stanford.edu/ng782rw8378 for details on dataset.
#
# The data are downloaded and organized locally into a BIDS compliant
# anatomical data folder (``anat``) and a diffusion-weighted imaging data
# (``dwi``) folder, which are both placed in the user's home directory under::
#
#   ``~/AFQ_data/stanford_hardi/``
#
# The data is also placed in a derivatives directory, signifying that it has
# already undergone the required preprocessing necessary for pyAFQ to run.
#
# The clear_previous_afq is used to remove any previous runs of the afq object
# stored in the `~/AFQ_data/stanford_hardi/` BIDS directory. Set it to None if
# you want to use the results of previous runs.

afd.organize_stanford_data(clear_previous_afq="track")

##########################################################################
# Set tractography parameters (optional)
# ---------------------------------------
# We make create a `tracking_params` variable, which we will pass to the
# GroupAFQ object which specifies that we want 25,000 seeds randomly
# distributed in the white matter. We only do this to make this example 
# faster and consume less space. We also set ``num_chunks`` to `True`,
# which will use ray to parallelize the tracking across all cores.
# This can be removed to process in serial, or set to use a particular
# distribution of work by setting `n_chunks` to an integer number.

tracking_params = dict(n_seeds=25000,
                       random_seeds=True,
                       rng_seed=2022,
                       trx=True,
                       num_chunks=True)

##########################################################################
# Initialize a GroupAFQ object:
# -------------------------
#
# Creates a GroupAFQ object, that encapsulates tractometry. This object can be
# used to manage the entire :doc:`AFQ pipeline`, including:
#
# - Tractography
# - Registration
# - Segmentation
# - Cleaning
# - Profiling
# - Visualization
#
# This will also create an output folder for the corresponding AFQ derivatives
# in the AFQ data directory: ``AFQ_data/stanford_hardi/derivatives/afq/``
#
# To initialize this object we will pass in the path location to our BIDS
# compliant data, the name of the preprocessing pipeline we want to use, and
# the tracking parameters we defined above. We will also specify the
# visualization backend we want to use (see below for more details).
# We will also be using plotly to generate an interactive visualization.
# The value `plotly_no_gif` indicates that interactive visualizations will be
# generated as html web-pages that can be opened in a browser, but not as
# static gif files.

myafq = GroupAFQ(
    bids_path=op.join(afd.afq_home, 'stanford_hardi'),
    preproc_pipeline='vistasoft',
    tracking_params=tracking_params,
    viz_backend_spec='plotly_no_gif')

##########################################################################
# Calculating DTI FA (Diffusion Tensor Imaging Fractional Anisotropy)
# ------------------------------------------------------------------
# The GroupAFQ object has a method called `export`, which allows the user
# to calculate various derived quantities from the data.
#
# For example, FA can be computed using the DTI model, by explicitly
# calling `myafq.export("dti_fa")`. This triggers the computation of DTI
# parameters for all subjects in the dataset, and stores the results in
# the AFQ derivatives directory. In addition, it calculates the FA
# from these parameters and stores it in a different file in the same
# directory.
#
# .. note::
#
#    The AFQ API computes quantities lazily. This means that DTI parameters
#    are not computed until they are required. This means that the first
#    line below is the one that requires time.
#
# The result of the call to `export` is a dictionary, with the subject
# IDs as keys, and the filenames of the corresponding files as values.
# This means that to extract the filename corresponding to the FA of the first
# subject, we can do:

FA_fname = myafq.export("dti_fa")["01"]

# We will then use `nibabel` to load the deriviative file and retrieve the
# data array.

FA_img = nib.load(FA_fname)
FA = FA_img.get_fdata()

##########################################################################
# Visualize the result with Matplotlib
# -------------------------------------
# At this point `FA` is an array, and we can use standard Python tools to
# visualize it or perform additional computations with it.
#
# In this case we are going to take an axial slice halfway through the
# FA data array and plot using a sequential color map.
#
# .. note::
#
#    The data array is structured as a xyz coordinate system.

fig, ax = plt.subplots(1)
ax.matshow(FA[:, :, FA.shape[-1] // 2], cmap='viridis')
ax.axis("off")

##########################################################################
# Recognizing the bundles and calculating tract profiles:
# -----------------------------------------------------
# Typically, users of pyAFQ are interested in calculating not only an overall
# map of the FA, but also the major white matter pathways (or bundles) and
# tract profiles of tissue properties along their length. To trigger the
# pyAFQ pipeline that calculates the profiles, users can call the
# `export('profiles')` method:
#
# .. note::
#    Running the code below triggers the full pipeline of operations
#    leading to the computation of the tract profiles. Therefore, it
#    takes a little while to run (about 40 minutes, typically).

myafq.export('profiles')

##########################################################################
# Visualizing the bundles and calculating act profiles:
# -----------------------------------------------------
# The pyAFQ API provides several ways to visualize bundles and profiles.
#
# First, we will run a function that exports an html file that contains
# an interactive visualization of the bundles that are segmented.
#
# .. note::
#    By default we resample a 100 points within a bundle, however to reduce
#    processing time we will only resample 50 points.
#
# Once it is done running, it should pop a browser window open and let you
# interact with the bundles.
#
# .. note::
#    You can hide or show a bundle by clicking the legend, or select a
#    single bundle by double clicking the legend. The interactive
#    visualization will also all you to pan, zoom, and rotate.

bundle_html = myafq.export("all_bundles_figure")
plotly.io.show(bundle_html["01"][0])

##########################################################################
# We can also visualize the tract profiles in all of the bundles. These
# plots show both FA (left) and MD (right) layed out anatomically.
# To make this plot, it is required that you install with
# `pip install pyAFQ[plot]` so that you have the necessary dependencies.
#

fig_files = myafq.export("tract_profile_plots")["01"]

##########################################################################
# .. figure:: {{ fig_files[0] }}
#


##########################################################################
# We can even use altair to visualize the tract profiles in all
# of the bundles. We provide a more customizable interface for visualizing
# the tract profiles using altair.
# Again, to make this plot, it is required that you install with
# `pip install pyAFQ[plot]` so that you have the necessary dependencies.
#
profiles_df = myafq.combine_profiles()
altair_df = ava.combined_profiles_df_to_altair_df(profiles_df)
altair_chart = ava.altair_df_to_chart(altair_df)
altair_chart.display()


##########################################################################
# We can check the number of streamlines per bundle, to make sure
# every bundle is found with a reasonable amount of streamlines.

bundle_counts = pd.read_csv(myafq.export("sl_counts")["01"], index_col=[0])
for ind in bundle_counts.index:
    if ind == "Total Recognized":
        threshold = 1000
    else:
        threshold = 10
    if bundle_counts["n_streamlines"][ind] < threshold:
        raise ValueError((
            "Small number of streamlines found "
            f"for bundle(s):\n{bundle_counts}"))
